# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pathlib
import subprocess
import tempfile
from typing import List, Optional

import gnupg

from snapcraft.internal.meta import package_repository

from . import apt_ppa, errors

logger = logging.getLogger(__name__)


class AptKeyManager:
    def __init__(
        self,
        *,
        gpg_keyring: pathlib.Path = pathlib.Path(
            "/etc/apt/trusted.gpg.d/snapcraft.gpg"
        ),
        key_assets: pathlib.Path,
    ) -> None:
        self._gpg_keyring = gpg_keyring
        self._key_assets = key_assets

    def find_asset_with_key_id(self, *, key_id: str) -> Optional[pathlib.Path]:
        """Find snap key asset matching key_id.

        The key asset much be named with the last 8 characters of the key
        identifier, in upper case.

        :param key_id: Key ID to search for.

        :returns: Path of key asset if match found, otherwise None.
        """
        key_file = key_id[-8:].upper() + ".asc"
        key_path = self._key_assets / key_file

        if key_path.exists():
            return key_path

        return None

    def get_key_fingerprints(self, *, key: str) -> List[str]:
        """List fingerprints found in specified key.

        Do this by importing the key into a temporary keyring,
        then querying the keyring for fingerprints.

        :param key: Key data (string) to parse.

        :returns: List of key fingerprints/IDs.
        """
        with tempfile.NamedTemporaryFile(suffix="keyring") as temp_file:
            return (
                gnupg.GPG(keyring=temp_file.name).import_keys(key_data=key).fingerprints
            )

    def is_key_installed(self, *, key_id: str) -> bool:
        """Check if specified key_id is installed.

        Check if key is installed by attempting to export the key.
        Unfortunately, apt-key does not exit with error and
        we have to do our best to parse the output.

        :param key_id: Key ID to check for.

        :returns: True if key is installed.
        """
        try:
            proc = subprocess.run(
                ["sudo", "apt-key", "export", key_id],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            # Export shouldn't exit with failure based on testing,
            # but assume the key is not installed and log a warning.
            logger.warning(f"Unexpected apt-key failure: {error.output}")
            return False

        apt_key_output = proc.stdout.decode()

        if "BEGIN PGP PUBLIC KEY BLOCK" in apt_key_output:
            return True

        if "nothing exported" in apt_key_output:
            return False

        # The two strings above have worked in testing, but if neither is
        # present for whatever reason, assume the key is not installed
        # and log a warning.
        logger.warning(f"Unexpected apt-key output: {apt_key_output}")
        return False

    def install_key(self, *, key: str) -> None:
        """Install given key.

        :param key: Key to install.

        :raises: AptGPGKeyInstallError if unable to install key.
        """
        cmd = [
            "sudo",
            "apt-key",
            "--keyring",
            str(self._gpg_keyring),
            "add",
            "-",
        ]

        try:
            logger.debug(f"Executing: {cmd!r}")
            env = dict()
            env["LANG"] = "C.UTF-8"
            subprocess.run(
                cmd,
                input=key.encode(),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
                env=env,
            )
        except subprocess.CalledProcessError as error:
            raise errors.AptGPGKeyInstallError(output=error.output.decode(), key=key)

        logger.debug(f"Installed apt repository key:\n{key}")

    def install_key_from_keyserver(
        self, *, key_id: str, key_server: str = "keyserver.ubuntu.com"
    ) -> None:
        """Install key from specified key server.

        :param key_id: Key ID to install.
        :param key_server: Key server to query.

        :raises: AptGPGKeyInstallError if unable to install key.
        """
        env = dict()
        env["LANG"] = "C.UTF-8"

        cmd = [
            "sudo",
            "apt-key",
            "--keyring",
            str(self._gpg_keyring),
            "adv",
            "--keyserver",
            key_server,
            "--recv-keys",
            key_id,
        ]

        try:
            logger.debug(f"Executing: {cmd!r}")
            subprocess.run(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
                env=env,
            )
        except subprocess.CalledProcessError as error:
            raise errors.AptGPGKeyInstallError(
                output=error.output.decode(), key_id=key_id, key_server=key_server
            )

    def install_package_repository_key(
        self, *, package_repo: package_repository.PackageRepository
    ) -> bool:
        """Install required key for specified package repository.

        For both PPA and other Apt package repositories:
        1) If key is already installed, return False.
        2) Install key from local asset, if available.
        3) Install key from key server, if available. An unspecified
           keyserver will default to using keyserver.ubuntu.com.

        :param package_repo: Apt PackageRepository configuration.

        :returns: True if key configuration was changed. False if
            key already installed.

        :raises: AptGPGKeyInstallError if unable to install key.
        """
        key_server: Optional[str] = None
        if isinstance(package_repo, package_repository.PackageRepositoryAptPpa):
            key_id = apt_ppa.get_launchpad_ppa_key_id(ppa=package_repo.ppa)
        elif isinstance(package_repo, package_repository.PackageRepositoryApt):
            key_id = package_repo.key_id
            key_server = package_repo.key_server
        else:
            raise RuntimeError(f"unhandled package repo type: {package_repo!r}")

        # Already installed, nothing to do.
        if self.is_key_installed(key_id=key_id):
            return False

        key_path = self.find_asset_with_key_id(key_id=key_id)
        if key_path is not None:
            self.install_key(key=key_path.read_text())
        else:
            if key_server is None:
                key_server = "keyserver.ubuntu.com"
            self.install_key_from_keyserver(key_id=key_id, key_server=key_server)

        return True



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shlex
from abc import ABC, abstractmethod
from subprocess import CalledProcessError
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from snapcraft import formatting_utils
from snapcraft.internal import steps

if TYPE_CHECKING:
    from snapcraft.internal.pluginhandler._dirty_report import DirtyReport
    from snapcraft.internal.pluginhandler._outdated_report import OutdatedReport


# Commonly used resolution message to clean and retry build.
CLEAN_RESOLUTION = "Run `snapcraft clean` and retry build."


def details_from_command_error(
    *,
    cmd: List[str],
    returncode: int,
    stdout: Optional[Union[bytes, str]] = None,
    stderr: Optional[Union[bytes, str]] = None,
) -> str:
    """Create a consistent details message from command errors.

    stdout and stderr, if provided, will be stringified using its object
    representation.  This method does not decode byte strings.

    :param cmd: Command executed.
    :param returncode: Command exit code.
    :param stdout: Optional stdout to include.
    :param stderr: Optional stderr to include.

    :returns: Details string.
    """
    cmd_string = " ".join([shlex.quote(c) for c in cmd])

    details = [
        f"* Command that failed: {cmd_string!r}",
        f"* Command exit code: {returncode}",
    ]

    if stdout:
        details.append(f"* Command output: {stdout!r}")

    if stderr:
        details.append(f"* Command standard error output: {stderr!r}")

    return "\n".join(details)


def details_from_called_process_error(error: CalledProcessError) -> str:
    """Create a consistent details message from CalledProcessError.

    :param error: CalledProcessError.

    :returns: Details string.
    """
    return details_from_command_error(
        cmd=error.cmd,
        stdout=error.stdout,
        stderr=error.stderr,
        returncode=error.returncode,
    )


class SnapcraftError(Exception):
    """DEPRECATED: Use SnapcraftException instead."""

    fmt = "Daughter classes should redefine this"

    def __init__(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return self.fmt.format([], **self.__dict__)

    def get_exit_code(self):
        """Exit code to use if this exception causes Snapcraft to exit."""
        return 2


class SnapcraftReportableError(SnapcraftError):
    """DEPRECATED: Use SnapcraftException instead with reportable=True."""


class SnapcraftException(Exception, ABC):
    """Base class for Snapcraft Exceptions."""

    @abstractmethod
    def get_brief(self) -> str:
        """Concise, single-line description of the error."""

    @abstractmethod
    def get_resolution(self) -> str:
        """Concise suggestion for user to resolve error."""

    def get_details(self) -> Optional[str]:
        """Detailed technical information, if required for user to debug issue."""
        return None

    def get_docs_url(self) -> Optional[str]:
        """Link to documentation on docs.snapcraft.io, if applicable."""
        return None

    def get_exit_code(self) -> int:
        """Exit code to use when exiting snapcraft due to this exception."""
        return 2

    def get_reportable(self) -> bool:
        """Defines if error is reportable (an exception trace should be shown)."""
        return False

    def __str__(self) -> str:
        return self.get_brief()


class SnapcraftReportableException(SnapcraftException, ABC):
    """Helper class for reportable Snapcraft Exceptions."""

    def get_reportable(self) -> bool:
        return True


class MissingStateCleanError(SnapcraftException):
    def __init__(self, step: steps.Step) -> None:
        self.step = step

    def get_brief(self) -> str:
        return f"Failed to clean for step {self.step.name!r}."

    def get_resolution(self) -> str:
        return CLEAN_RESOLUTION


class StepOutdatedError(SnapcraftException):
    def __init__(
        self,
        *,
        step: steps.Step,
        part: str,
        dirty_report: "DirtyReport" = None,
        outdated_report: "OutdatedReport" = None,
    ) -> None:
        self.step = step
        self.part = part
        self.parts_names = part
        self.dirty_report = dirty_report
        self.outdated_report = outdated_report
        self.report = self._consolidate_reports()

    def _consolidate_reports(self) -> str:
        messages = []

        if self.dirty_report:
            messages.append(self.dirty_report.get_report())

        if self.outdated_report:
            messages.append(self.outdated_report.get_report())

        return "".join(messages)

    def get_brief(self) -> str:
        return "Failed to reuse files from previous run."

    def get_resolution(self) -> str:
        return CLEAN_RESOLUTION

    def get_details(self) -> Optional[str]:
        # If non-empty string, return report (otherwise None).
        if self.report:
            return self.report

        return None


class SnapcraftEnvironmentError(SnapcraftException):
    """DEPRECATED: Too generic, create (or re-use) a tailored one."""

    # FIXME This exception is too generic.
    # https://bugs.launchpad.net/snapcraft/+bug/1734231
    # --elopio - 20171123

    def __init__(self, message: str) -> None:
        self.message = message

    def get_brief(self) -> str:
        return self.message

    def get_resolution(self) -> str:
        return ""


class SnapcraftDataDirectoryMissingError(SnapcraftReportableException):
    def get_brief(self) -> str:
        return "Cannot find snapcraft's data files."

    def get_resolution(self) -> str:
        return "Re-install snapcraft or verify installation is correct."


class SnapcraftMissingLinkerInBaseError(SnapcraftException):
    def __init__(self, *, base: str, linker_path: str) -> None:
        self.base = base
        self.linker_path = linker_path

    def get_brief(self) -> str:
        return f"Cannot find the linker to use for the target base {self.base!r}."

    def get_resolution(self) -> str:
        return (
            f"Verify that the linker exists at the expected path "
            f"{self.linker_path!r} and try again. If the linker "
            f"does not exist contact the author of the base (run "
            f"`snap info {self.base}` to get information for this base)."
        )


class IncompatibleBaseError(SnapcraftError):
    fmt = (
        "The linker version {linker_version!r} used by the base {base!r} is "
        "incompatible with files in this snap:\n"
        "{file_list}"
    )

    def __init__(
        self, *, base: str, linker_version: str, file_list: Dict[str, str]
    ) -> None:
        spaced_file_list = ("    {} ({})".format(k, v) for k, v in file_list.items())
        super().__init__(
            base=base,
            linker_version=linker_version,
            file_list="\n".join(sorted(spaced_file_list)),
        )


class PrimeFileConflictError(SnapcraftError):

    fmt = (
        "Failed to filter files: "
        "The following files have been excluded by the `stage` keyword, "
        "but included by the `prime` keyword: {fileset!r}. "
        "Edit the `snapcraft.yaml` to make sure that the files included in "
        "`prime` are also included in `stage`."
    )


class PluginError(SnapcraftError):

    fmt = (
        "Failed to load plugin: "
        "{message}"
        # FIXME include how to fix each of the possible plugin errors.
        # https://bugs.launchpad.net/snapcraft/+bug/1727484
        # --elopio - 2017-10-25
    )

    def __init__(self, message):
        super().__init__(message=message)


class PluginBaseError(SnapcraftError):

    fmt = "The plugin used by part {part_name!r} does not support snaps using base {base!r}."

    def __init__(self, *, part_name, base):
        super().__init__(part_name=part_name, base=base)


class SnapcraftPartConflictError(SnapcraftError):

    fmt = (
        "Failed to stage: "
        "Parts {other_part_name!r} and {part_name!r} have the following "
        "files, but with different contents:\n"
        "{file_paths}\n\n"
        "Snapcraft offers some capabilities to solve this by use of the "
        "following keywords:\n"
        "    - `filesets`\n"
        "    - `stage`\n"
        "    - `snap`\n"
        "    - `organize`\n\n"
        "To learn more about these part keywords, run "
        "`snapcraft help plugins`."
    )

    def __init__(self, *, part_name, other_part_name, conflict_files):
        spaced_conflict_files = ("    {}".format(i) for i in conflict_files)
        super().__init__(
            part_name=part_name,
            other_part_name=other_part_name,
            file_paths="\n".join(sorted(spaced_conflict_files)),
        )


class SnapcraftOrganizeError(SnapcraftError):

    fmt = "Failed to organize part {part_name!r}: {message}"

    def __init__(self, part_name, message):
        super().__init__(part_name=part_name, message=message)


class InvalidWikiEntryError(SnapcraftError):

    fmt = (
        "Invalid wiki entry: "
        "{error!r}"
        # FIXME include how to fix each of the possible wiki errors.
        # https://bugs.launchpad.net/snapcraft/+bug/1727490
        # --elopio - 2017-10-25
    )

    def __init__(self, error=None):
        super().__init__(error=error)


class MissingGadgetError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "Missing gadget.yaml file.\n"
        "When creating gadget snaps you are required to provide a gadget.yaml "
        "file\n"
        "in the root of your snapcraft project\n\n"
        "Read more about gadget snaps and the gadget.yaml on:\n"
        "https://github.com/snapcore/snapd/wiki/Gadget-snap"
    )


class PluginOutdatedError(SnapcraftError):

    fmt = "This plugin is outdated: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class ToolMissingError(SnapcraftReportableError):

    fmt = (
        "A tool snapcraft depends on could not be found: {command_name!r}.\n"
        "Ensure the tool is installed and available, and try again."
    )

    def __init__(self, *, command_name: str) -> None:
        super().__init__(command_name=command_name)


class RequiredCommandFailure(SnapcraftError):

    fmt = "{command!r} failed."


class RequiredCommandNotFound(SnapcraftError):

    fmt = "{cmd_list[0]!r} not found."


class RequiredPathDoesNotExist(SnapcraftError):

    fmt = "Required path does not exist: {path!r}"


class SnapcraftPathEntryError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "The path {value!r} set for {key!r} in {app!r} does not exist. "
        "Make sure that the files are in the `prime` directory."
    )


class InvalidPullPropertiesError(SnapcraftError):

    fmt = (
        "Failed to load plugin: "
        "Invalid pull properties specified by {plugin_name!r} plugin: "
        "{properties}"
        # TODO add a link to some docs about pull properties?
        # --elopio - 2017-10-25
    )

    def __init__(self, plugin_name, properties):
        super().__init__(plugin_name=plugin_name, properties=properties)


class InvalidBuildPropertiesError(SnapcraftError):

    fmt = (
        "Failed to load plugin: "
        "Invalid build properties specified by {plugin_name!r} plugin: "
        "{properties}"
        # TODO add a link to some docs about build properties?
        # --elopio - 2017-10-25
    )

    def __init__(self, plugin_name, properties):
        super().__init__(plugin_name=plugin_name, properties=properties)


class StagePackageDownloadError(SnapcraftError):

    fmt = (
        "Failed to fetch stage packages: "
        "Error downloading packages for part {part_name!r}: {message}."
    )

    def __init__(self, part_name, message):
        super().__init__(part_name=part_name, message=message)


class OsReleaseIdError(SnapcraftError):

    fmt = "Unable to determine host OS ID"


class OsReleaseNameError(SnapcraftError):

    fmt = "Unable to determine host OS name"


class OsReleaseVersionIdError(SnapcraftError):

    fmt = "Unable to determine host OS version ID"


class OsReleaseCodenameError(SnapcraftError):

    fmt = "Unable to determine host OS version codename"


class InvalidContainerImageInfoError(SnapcraftError):

    fmt = (
        "Failed to parse container image info: "
        "SNAPCRAFT_IMAGE_INFO is not a valid JSON string: {image_info}"
    )

    def __init__(self, image_info: str) -> None:
        super().__init__(image_info=image_info)


class PatcherError(SnapcraftError):
    pass


class PatcherGenericError(PatcherError):

    fmt = (
        "{elf_file!r} cannot be patched to function properly in a classic "
        "confined snap: {message}"
    )

    def __init__(self, *, elf_file, process_exception):
        message = "{} failed with exit code {}".format(
            " ".join(process_exception.cmd), process_exception.returncode
        )
        super().__init__(elf_file=elf_file, message=message)


class PatcherNewerPatchelfError(PatcherError):

    fmt = (
        "{elf_file!r} cannot be patched to function properly in a classic "
        "confined snap: {message}.\n"
        "{patchelf_version!r} may be too old. A newer version of patchelf "
        "may be required.\n"
        "Try adding the `after: [patchelf]` and a `patchelf` part that would "
        "filter out files from prime `prime: [-*]` or "
        "`build-snaps: [patchelf/latest/edge]` to the failing part in your "
        "`snapcraft.yaml` to use a newer patchelf."
    )

    def __init__(self, *, elf_file, process_exception, patchelf_version):
        message = "{} failed with exit code {}".format(
            " ".join(process_exception.cmd), process_exception.returncode
        )
        super().__init__(
            elf_file=elf_file, message=message, patchelf_version=patchelf_version
        )


class StagePackageMissingError(SnapcraftError):

    fmt = (
        "{package!r} is required inside the snap for this part to work "
        "properly.\n"
        "Add it as a `stage-packages` entry for this part."
    )

    def __init__(self, *, package):
        super().__init__(package=package)


class MetadataExtractionError(SnapcraftError):
    pass


class MissingMetadataFileError(MetadataExtractionError):

    fmt = (
        "Failed to generate snap metadata: "
        "Part {part_name!r} has a 'parse-info' referring to metadata file "
        "{path!r}, which does not exist."
    )

    def __init__(self, part_name: str, path: str) -> None:
        super().__init__(part_name=part_name, path=path)


class UnhandledMetadataFileTypeError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "This type of file is not supported for supplying metadata."
    )

    def __init__(self, path: str) -> None:
        super().__init__(path=path)


class InvalidExtractorValueError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "Extractor {extractor_name!r} didn't return ExtractedMetadata as "
        "expected."
    )

    def __init__(self, path: str, extractor_name: str) -> None:
        super().__init__(path=path, extractor_name=extractor_name)


class SnapcraftCommandError(SnapcraftError, CalledProcessError):
    """Exception for generic command errors.

    Processes should capture this error for specific messaging.
    This exception carries the signature of CalledProcessError for backwards
    compatibility.
    """

    fmt = "Failed to run {command!r}: Exited with code {exit_code}."

    def __init__(self, *, command: str, call_error: CalledProcessError) -> None:
        super().__init__(command=command, exit_code=call_error.returncode)
        CalledProcessError.__init__(
            self,
            returncode=call_error.returncode,
            cmd=call_error.cmd,
            output=call_error.output,
            stderr=call_error.stderr,
        )


class SnapcraftPluginCommandError(SnapcraftError):
    """Command executed by a plugin fails."""

    fmt = (
        "Failed to run {command!r} for {part_name!r}: "
        "Exited with code {exit_code}.\n"
        "Verify that the part is using the correct parameters and try again."
    )

    def __init__(
        self, *, command: Union[List, str], part_name: str, exit_code: int
    ) -> None:
        if isinstance(command, list):
            command = " ".join([shlex.quote(c) for c in command])
        super().__init__(command=command, part_name=part_name, exit_code=exit_code)


class ScriptletBaseError(SnapcraftError):
    """Base class for all scriptlet-related exceptions.

    :cvar fmt: A format string that daughter classes override

    """


class ScriptletRunError(ScriptletBaseError):
    fmt = "Failed to run {scriptlet_name!r}: Exit code was {code}."

    def __init__(self, scriptlet_name: str, code: int) -> None:
        super().__init__(scriptlet_name=scriptlet_name, code=code)


class ScriptletDuplicateDataError(ScriptletBaseError):
    fmt = (
        "Failed to save data from scriptlet into {step.name!r} state: "
        "The {humanized_keys} key(s) were already saved in the "
        "{other_step.name!r} step."
    )

    def __init__(
        self, step: steps.Step, other_step: steps.Step, keys: List[str]
    ) -> None:
        self.keys = keys
        super().__init__(
            step=step,
            other_step=other_step,
            humanized_keys=formatting_utils.humanize_list(keys, "and"),
        )


class ScriptletDuplicateFieldError(ScriptletBaseError):
    fmt = "Unable to set {field}: it was already set in the {step.name!r} step."

    def __init__(self, field: str, step: steps.Step) -> None:
        super().__init__(field=field, step=step)


class SnapcraftctlError(ScriptletBaseError):
    fmt = "{message}"

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class SnapcraftInvalidCLIConfigError(SnapcraftError):

    fmt = "The cli configuration file {config_file!r} has invalid data: {error!r}."

    def __init__(self, *, config_file: str, error: str) -> None:
        super().__init__(config_file=config_file, error=error)
        # This is to keep mypy happy
        self.config_file = config_file


class SnapcraftCopyFileNotFoundError(SnapcraftError):
    fmt = (
        "Failed to copy {path!r}: no such file or directory.\n"
        "Check the path and try again."
    )

    def __init__(self, path):
        super().__init__(path=path)


class InvalidStepError(SnapcraftError):
    fmt = "{step_name!r} is not a valid lifecycle step"

    def __init__(self, step_name):
        super().__init__(step_name=step_name)


class StepHasNotRunError(SnapcraftError):
    fmt = "The {part_name!r} part has not yet run the {step.name!r} step"

    def __init__(self, part_name, step):
        super().__init__(part_name=part_name, step=step)


class NoLatestStepError(SnapcraftError):
    fmt = "The {part_name!r} part hasn't run any steps"

    def __init__(self, part_name):
        super().__init__(part_name=part_name)


class NoNextStepError(SnapcraftError):
    fmt = "The {part_name!r} part has run through its entire lifecycle"

    def __init__(self, part_name):
        super().__init__(part_name=part_name)


class MountPointNotFoundError(SnapcraftError):
    fmt = "Nothing is mounted at {mount_point!r}"

    def __init__(self, mount_point):
        super().__init__(mount_point=mount_point)


class RootNotMountedError(SnapcraftError):
    fmt = "{root!r} is not mounted"

    def __init__(self, root):
        super().__init__(root=root)


class InvalidMountinfoFormat(SnapcraftError):
    fmt = "Unable to parse mountinfo row: {row}"

    def __init__(self, row):
        super().__init__(row=row)


class CrossCompilationNotSupported(SnapcraftError):
    fmt = (
        "The plugin used by {part_name!r} does not support cross-compiling "
        "to a different target architecture."
    )

    def __init__(self, *, part_name: str) -> None:
        super().__init__(part_name=part_name)


class SnapDataExtractionError(SnapcraftError):
    fmt = "Cannot read data from snap {snap!r}. The file may be corrupted."

    def __init__(self, snap):
        super().__init__(snap=snap)


class ProjectNotFoundError(SnapcraftReportableError):
    fmt = "Failed to find project files."


class XAttributeError(SnapcraftReportableException):
    def __init__(self, *, action: str, key: str, path: str) -> None:
        self._action = action
        self._key = key
        self._path = path

    def get_brief(self) -> str:
        return f"Unable to {self._action} extended attribute."

    def get_details(self) -> str:
        return f"Failed to {self._action} attribute {self._key!r} on {self._path!r}."

    def get_resolution(self) -> str:
        return "Check that your filesystem supports extended attributes."


class XAttributeTooLongError(SnapcraftReportableException):
    def __init__(self, *, key: str, value: str, path: str) -> None:
        self._key = key
        self._value = value
        self._path = path

    def get_brief(self) -> str:
        return "Unable to write extended attribute as the key and/or value is too long."

    def get_details(self) -> str:
        return (
            f"Failed to write attribute to {self._path!r}:\n"
            f"key={self._key!r} value={self._value!r}"
        )

    def get_resolution(self) -> str:
        return (
            "This issue is generally resolved by addressing/truncating "
            "the data source of the long data value. In some cases, the "
            "filesystem being used will limit the allowable size."
        )


class CorruptedElfFileError(SnapcraftException):
    def __init__(self, path: str, error: Exception) -> None:
        self.path = path
        self.message = str(error)

    def get_brief(self) -> str:
        return f"Unable to parse ELF file {self.path!r}: {self.message}"

    def get_resolution(self) -> str:
        return "Remove the suspect files from the snap using the `stage` or `prime` keywords."


class SnapcraftPluginAssertionError(SnapcraftException):
    def __init__(self, *, name: str, reason: str, details: str = "") -> None:
        self._name = name
        self._reason = reason
        self._details = details

    def get_brief(self) -> str:
        return f"Unable to build {self._name!r}: {self._reason}"

    def get_details(self) -> str:
        return self._details

    def get_resolution(self) -> str:
        return "Ensure the part's configuration and sources are correct."


class SnapcraftPluginBuildError(SnapcraftException):
    """An exception to raise when the PluginV2 build fails at runtime."""

    def __init__(self, *, part_name: str) -> None:
        self._part_name = part_name

    def get_brief(self) -> str:
        return f"Failed to build {self._part_name!r}."

    def get_resolution(self) -> str:
        return "Check the build logs and ensure the part's configuration and sources are correct."


class SnapcraftHostToolNotFoundError(SnapcraftException):
    """An exception to raise when a host tool is required, but not found."""

    def __init__(self, *, command_name: str, package_name: str) -> None:
        self._command_name = command_name
        self._package_name = package_name

    def get_brief(self) -> str:
        return f"A tool snapcraft depends on could not be found: {self._command_name!r}"

    def get_resolution(self) -> str:
        return f"Ensure that {self._package_name!r} is installed."


class BuildAttributePatchelfConflictError(SnapcraftException):
    """An exception to raise a part specified incompatible patchelf-related build-attributes."""

    def __init__(self, *, part_name: str) -> None:
        self._part_name = part_name

    def get_brief(self) -> str:
        return f"part {self._part_name} has both 'no-patchelf' and 'enable-patchelf' build-attributes, which are mutually exclusive and cannot be used together."

    def get_resolution(self) -> str:
        return f"Use either 'no-patchelf' or 'enable-patchelf', not both."



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil

from snapcraft.file_utils import calculate_hash


class TestFileCache:

    scenarios = [
        ("sha384", dict(algo="sha384")),
        ("md5", dict(algo="md5")),
        ("sha1", dict(algo="sha1")),
        ("sha224", dict(algo="sha224")),
        ("sha256", dict(algo="sha256")),
        ("sha384", dict(algo="sha384")),
        ("sha512", dict(algo="sha512")),
        ("sha3_256", dict(algo="sha3_256")),
        ("sha3_384", dict(algo="sha3_384")),
        ("sha3_512", dict(algo="sha3_512")),
    ]

    def test_get_nothing_cached(self, file_cache, algo):
        assert file_cache.get(algorithm=algo, hash="1") is None

    def test_cache_and_retrieve(self, random_data_file, file_cache, algo):
        calculated_hash = calculate_hash(random_data_file, algorithm=algo)
        cached_file = file_cache.cache(
            filename=random_data_file, algorithm=algo, hash=calculated_hash
        )
        leaf_path = os.path.join(algo, calculated_hash)
        assert cached_file.endswith(leaf_path)

        retrieved_file = file_cache.get(algorithm=algo, hash=calculated_hash)
        assert retrieved_file.endswith(leaf_path)

    def test_cache_not_possible(self, random_data_file, file_cache, algo):
        bad_calculated_hash = "1"

        cached_file = file_cache.cache(
            filename=random_data_file, algorithm=algo, hash=bad_calculated_hash
        )
        assert cached_file is None

    def test_cache_file_copy_error(
        self, monkeypatch, random_data_file, file_cache, algo
    ):
        calculated_hash = calculate_hash(random_data_file, algorithm=algo)

        def fake_copy(*args, **kwargs):
            raise OSError()

        monkeypatch.setattr(shutil, "copyfile", fake_copy)

        cached_file = file_cache.cache(
            filename=random_data_file, algorithm=algo, hash=calculated_hash
        )

        assert cached_file is None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals

import snapcraft
from snapcraft.internal.remote_build import LaunchpadClient, errors
from snapcraft.internal.sources._git import Git
from snapcraft.internal.sources.errors import SnapcraftPullError
from tests import unit

from . import TestDir


class FakeLaunchpadObject:
    """Mimic behavior of many launchpad objects."""

    def __init__(self):
        pass

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return self.__getattribute__(key)


class BuildImpl(FakeLaunchpadObject):
    def __init__(self, fake_arch="i386"):
        self._fake_arch = fake_arch
        self.getFileUrls_mock = mock.Mock(
            return_value=[f"url_for/snap_file_{self._fake_arch}.snap"]
        )

    def getFileUrls(self, *args, **kw):
        return self.getFileUrls_mock(*args, **kw)


class SnapBuildEntryImpl(FakeLaunchpadObject):
    def __init__(self, arch_tag="", buildstate="", self_link="", build_log_url=""):
        self.arch_tag = arch_tag
        self.buildstate = buildstate
        self.self_link = self_link
        self.build_log_url = build_log_url


class SnapBuildsImpl(FakeLaunchpadObject):
    def __init__(
        self,
        entries=[
            SnapBuildEntryImpl(
                arch_tag="i386",
                buildstate="Successfully built",
                self_link="http://build_self_link_1",
                build_log_url="url_for/build_log_file_1",
            ),
            SnapBuildEntryImpl(
                arch_tag="amd64",
                buildstate="Failed to build",
                self_link="http://build_self_link_2",
                build_log_url="url_for/build_log_file_2",
            ),
            SnapBuildEntryImpl(
                arch_tag="arm64",
                buildstate="Failed to build",
                self_link="http://build_self_link_2",
                build_log_url=None,
            ),
        ],
    ):
        self.entries = entries


class SnapBuildReqImpl(FakeLaunchpadObject):
    def __init__(
        self,
        status="Completed",
        error_message="",
        self_link="http://request_self_link/1234",
        builds_collection_link="http://builds_collection_link",
    ):
        self.status = status
        self.error_message = error_message
        self.self_link = self_link
        self.builds_collection_link = builds_collection_link
        self.builds = SnapBuildsImpl()

    def lp_refresh(self):
        pass


class SnapImpl(FakeLaunchpadObject):
    def __init__(self, builds_collection_link="http://builds_collection_link"):
        self._req = SnapBuildReqImpl()
        self.lp_delete_mock = mock.Mock()
        self.requestBuilds_mock = mock.Mock(return_value=self._req)
        self.builds_collection_link = builds_collection_link

    def lp_delete(self, *args, **kw):
        return self.lp_delete_mock(*args, **kw)

    def requestBuilds(self, *args, **kw):
        return self.requestBuilds_mock(*args, **kw)


class SnapsImpl(FakeLaunchpadObject):
    def __init__(self):
        self._snap = SnapImpl()
        self.getByName_mock = mock.Mock(return_value=self._snap)
        self.new_mock = mock.Mock(return_value=self._snap)

    def getByName(self, *args, **kw):
        return self.getByName_mock(*args, **kw)

    def new(self, *args, **kw):
        return self.new_mock(*args, **kw)


class GitImpl(FakeLaunchpadObject):
    def __init__(self):
        self.issueAccessToken_mock = mock.Mock(return_value="access-token")
        self.lp_delete_mock = mock.Mock()

    def issueAccessToken(self, *args, **kw):
        return self.issueAccessToken_mock(*args, **kw)

    def lp_delete(self, *args, **kw):
        return self.lp_delete_mock(*args, **kw)


class GitRepositoriesImpl(FakeLaunchpadObject):
    def __init__(self):
        self._git = GitImpl()
        self.new_mock = mock.Mock(return_value=self._git)
        self.getByPath_mock = mock.Mock(return_value=self._git)

    def getByPath(self, *args, **kw):
        return self.getByPath_mock(*args, **kw)

    def new(self, *args, **kw):
        return self.new_mock(*args, **kw)


class DistImpl(FakeLaunchpadObject):
    def __init__(self):
        self.main_archive = "main_archive"


class MeImpl(FakeLaunchpadObject):
    def __init__(self):
        self.name = "user"


class LaunchpadImpl(FakeLaunchpadObject):
    def __init__(self):
        self._login_mock = mock.Mock()
        self._load_mock = mock.Mock()
        self._rbi = SnapBuildReqImpl()

        self.git_repositories = GitRepositoriesImpl()
        self.snaps = SnapsImpl()
        self.people = {"user": "/~user"}
        self.distributions = {"ubuntu": DistImpl()}
        self.rbi = self._rbi
        self.me = MeImpl()

    def load(self, url: str, *args, **kw):
        self._load_mock(url, *args, **kw)
        if "/+build-request/" in url:
            return self._rbi
        elif "http://build_self_link_1" in url:
            return BuildImpl(fake_arch="i386")
        elif "http://build_self_link_2" in url:
            return BuildImpl(fake_arch="amd64")
        else:
            return self._rbi.builds


class LaunchpadTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self._project = self._make_snapcraft_project()
        self.lp = LaunchpadImpl()
        self.fake_login_with = fixtures.MockPatch(
            "launchpadlib.launchpad.Launchpad.login_with", return_value=self.lp
        )
        self.useFixture(self.fake_login_with)

        self.mock_git_class = mock.Mock(spec=Git)
        self.mock_git_class.check_command_installed.return_value = True

        self.lpc = LaunchpadClient(
            project=self._project,
            build_id="id",
            architectures=[],
            git_class=self.mock_git_class,
            running_snapcraft_version="100.1234",
        )

    def test_login(self):
        self.assertThat(self.lpc.user, Equals("user"))
        self.fake_login_with.mock.assert_called_with(
            "snapcraft remote-build {}".format(snapcraft.__version__),
            "production",
            mock.ANY,
            credentials_file=mock.ANY,
            version="devel",
        )

    def test_login_connection_issues(self):
        self.fake_login_with.mock.side_effect = ConnectionRefusedError()
        self.assertRaises(errors.LaunchpadHttpsError, self.lpc.login)

        self.fake_login_with.mock.side_effect = TimeoutError()
        self.assertRaises(errors.LaunchpadHttpsError, self.lpc.login)

    def test_load_connection_refused(self):
        # ConnectionRefusedError should surface.
        self.fake_login_with.mock.reset_mock()
        self.lpc._lp._load_mock.side_effect = ConnectionRefusedError
        self.assertRaises(ConnectionRefusedError, self.lpc._lp_load_url, "foo")
        self.fake_login_with.mock.assert_not_called()

    def test_load_connection_reset_once(self):
        # Load URL should work OK after single connection reset.
        self.fake_login_with.mock.reset_mock()
        self.lpc._lp._load_mock.side_effect = [ConnectionResetError, None]
        self.lpc._lp_load_url(url="foo")
        self.fake_login_with.mock.assert_called()

    def test_load_connection_reset_twice(self):
        # Load URL should fail with two connection resets.
        self.fake_login_with.mock.reset_mock()
        self.lpc._lp._load_mock.side_effect = [
            ConnectionResetError,
            ConnectionResetError,
        ]
        self.assertRaises(ConnectionResetError, self.lpc._lp_load_url, "foo")
        self.fake_login_with.mock.assert_called()

    def test_create_snap(self):
        self.lpc._create_snap()
        self.lpc._lp.snaps.new_mock.assert_called_with(
            auto_build=False,
            auto_build_archive="/ubuntu/+archive/primary",
            auto_build_pocket="Updates",
            git_path="master",
            git_repository_url="https://user@git.launchpad.net/~user/+git/id/",
            name="id",
            owner="/~user",
        )

    def test_create_snap_with_archs(self):
        self.lpc.architectures = ["arch1", "arch2"]
        self.lpc._create_snap()
        self.lpc._lp.snaps.new_mock.assert_called_with(
            auto_build=False,
            auto_build_archive="/ubuntu/+archive/primary",
            auto_build_pocket="Updates",
            git_path="master",
            git_repository_url="https://user@git.launchpad.net/~user/+git/id/",
            name="id",
            owner="/~user",
            processors=["/+processors/arch1", "/+processors/arch2"],
        )

    def test_delete_snap(self):
        self.lpc._delete_snap()
        self.lpc._lp.snaps.getByName_mock.assert_called_with(name="id", owner="/~user")

    def test_start_build(self):
        self.lpc.start_build()

    @mock.patch(
        "tests.unit.remote_build.test_launchpad.SnapImpl.requestBuilds",
        return_value=SnapBuildReqImpl(
            status="Failed", error_message="snapcraft.yaml not found..."
        ),
    )
    def test_start_build_error(self, mock_rb):
        raised = self.assertRaises(errors.RemoteBuilderError, self.lpc.start_build)
        self.assertThat(str(raised), Contains("snapcraft.yaml not found..."))

    @mock.patch(
        "tests.unit.remote_build.test_launchpad.SnapImpl.requestBuilds",
        return_value=SnapBuildReqImpl(status="Pending", error_message=""),
    )
    @mock.patch("time.time", return_value=500)
    def test_start_build_error_timeout(self, mock_time, mock_rb):
        self.lpc.deadline = 499
        raised = self.assertRaises(errors.RemoteBuildTimeoutError, self.lpc.start_build)
        self.assertThat(
            str(raised), Equals("Remote build exceeded configured timeout.")
        )

    def test_issue_build_request_defaults(self):
        fake_snap = mock.MagicMock()

        self.lpc._issue_build_request(fake_snap)

        self.assertThat(
            fake_snap.mock_calls,
            Equals(
                [
                    mock.call.requestBuilds(
                        archive="main_archive",
                        channels={"core18": "stable", "snapcraft": "stable"},
                        pocket="Updates",
                    )
                ]
            ),
        )

    @mock.patch("snapcraft.internal.remote_build.LaunchpadClient._download_file")
    def test_monitor_build(self, mock_download_file):
        open("test_i386.txt", "w").close()
        open("test_i386.1.txt", "w").close()

        self.lpc.start_build()
        self.lpc.monitor_build(interval=0)
        self.assertThat(
            mock_download_file.mock_calls,
            Equals(
                [
                    mock.call(
                        dst="snap_file_i386.snap", url="url_for/snap_file_i386.snap"
                    ),
                    mock.call(
                        dst="test_i386.2.txt",
                        gunzip=True,
                        url="url_for/build_log_file_1",
                    ),
                    mock.call(
                        dst="snap_file_amd64.snap", url="url_for/snap_file_amd64.snap"
                    ),
                    mock.call(
                        dst="test_amd64.txt",
                        gunzip=True,
                        url="url_for/build_log_file_2",
                    ),
                    mock.call(
                        dst="snap_file_amd64.snap", url="url_for/snap_file_amd64.snap"
                    ),
                ]
            ),
        )

    @mock.patch("snapcraft.internal.remote_build.LaunchpadClient._download_file")
    @mock.patch(
        "tests.unit.remote_build.test_launchpad.BuildImpl.getFileUrls", return_value=[]
    )
    @mock.patch("logging.Logger.error")
    def test_monitor_build_error(self, mock_log, mock_urls, mock_download_file):
        self.lpc.start_build()
        self.lpc.monitor_build(interval=0)
        mock_download_file.assert_has_calls(
            [
                mock.call(
                    url="url_for/build_log_file_2", gunzip=True, dst="test_amd64.txt"
                )
            ]
        )
        self.assertThat(
            mock_log.mock_calls,
            Equals(
                [
                    mock.call("Snap file not available for arch 'i386'."),
                    mock.call("Snap file not available for arch 'amd64'."),
                    mock.call("Build failed for arch 'amd64'."),
                    mock.call("Snap file not available for arch 'arm64'."),
                    mock.call("Build failed for arch 'arm64'."),
                ]
            ),
        )

    @mock.patch("snapcraft.internal.remote_build.LaunchpadClient._download_file")
    @mock.patch("time.time", return_value=500)
    def test_monitor_build_error_timeout(self, mock_time, mock_rb):
        self.lpc.deadline = 499
        self.lpc.start_build()
        raised = self.assertRaises(
            errors.RemoteBuildTimeoutError, self.lpc.monitor_build, interval=0
        )
        self.assertThat(
            str(raised), Equals("Remote build exceeded configured timeout.")
        )

    def test_get_build_status(self):
        self.lpc.start_build()
        build_status = self.lpc.get_build_status()
        self.assertThat(
            build_status,
            Equals(
                {
                    "amd64": "Failed to build",
                    "arm64": "Failed to build",
                    "i386": "Successfully built",
                }
            ),
        )

    def _make_snapcraft_project(self):
        yaml = textwrap.dedent(
            """\
            name: test
            base: core18
            version: "1.0"
            summary: test
            description: test
            confinement: strict
            grade: stable
            """
        )
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(yaml)
        project = snapcraft.project.Project(
            snapcraft_yaml_file_path=snapcraft_yaml_file_path
        )
        return project

    def test_git_repository_creation(self):
        source_testdir = self.useFixture(TestDir())
        source_testdir.create_file("foo")
        repo_dir = source_testdir.path

        self.lpc._gitify_repository(repo_dir)

        assert self.mock_git_class.mock_calls == [
            mock.call.check_command_installed(),
            mock.call(repo_dir, repo_dir, silent=True),
            mock.call().init(),
            mock.call().add("foo"),
            mock.call().commit("committed by snapcraft version: 100.1234"),
        ]

    def test_push_source_tree(self):
        source_testdir = self.useFixture(TestDir())
        repo_dir = source_testdir.path

        self.lpc.push_source_tree(repo_dir)

        self.mock_git_class.assert_has_calls == [
            mock.call(
                "https://user:access-token@git.launchpad.net/~user/+git/id/",
                "HEAD:master",
                force=True,
            )
        ]

    def test_push_source_tree_error(self):
        self.mock_git_class.return_value.push.side_effect = (
            SnapcraftPullError(
                command="git push HEAD:master https://user:access-token@url",
                exit_code=128,
            ),
        )
        source_testdir = self.useFixture(TestDir())
        repo_dir = source_testdir.path

        self.assertRaises(
            errors.LaunchpadGitPushError, self.lpc.push_source_tree, repo_dir
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import random
from unittest import mock

import fixtures
from progressbar import AnimatedMarker, ProgressBar
from testtools import matchers as m

from snapcraft.internal import deltas
from tests import fixture_setup, unit


class XDelta3TestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(fixture_setup.FakeTerminal())
        self.fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(self.fake_logger)

        self.workdir = self.useFixture(fixtures.TempDir()).path
        self.source_file = os.path.join(self.workdir, "source.snap")
        self.target_file = os.path.join(self.workdir, "target.snap")

        with open(self.source_file, "wb") as f:
            f.write(b"This is the source file.")
        with open(self.target_file, "wb") as f:
            f.write(b"This is the target file.")

    def generate_snap_pair(self):
        """Generate more realistic snap data.

        Most tests don't need realistic data, and can use the default dummy
        data that's created by setUp. However, tests that actually run xdelta3
        could use more accurate data. This file generates larger
        binary files. The files will be similar with roughly 0.1 variance along
        1KB block boundaries.
        """
        mean_size = 2 ** 20
        size_stddev = 2 ** 15
        snap_size = int(random.normalvariate(mean_size, size_stddev))

        scratchdir = self.useFixture(fixtures.TempDir()).path
        self.source_file = os.path.join(scratchdir, "source-snap")
        self.target_file = os.path.join(scratchdir, "target-snap")
        # source snap is completely random:
        with open(self.source_file, "wb") as source, open(
            self.target_file, "wb"
        ) as target:
            for i in range(0, snap_size, 1024):
                block = os.urandom(1024)
                source.write(block)
                if random.randint(0, 9) == 0:
                    target.write(os.urandom(1024))
                else:
                    target.write(block)

    def test_xdelta3(self):
        self.generate_snap_pair()
        base_delta = deltas.XDelta3Generator(
            source_path=self.source_file, target_path=self.target_file
        )
        path = base_delta.make_delta(is_for_test=True)

        self.assertThat(path, m.FileExists())
        expect_path = "{}.{}".format(
            base_delta.target_path, base_delta.delta_file_extname
        )
        self.assertThat(path, m.Equals(expect_path))

    def test_xdelta3_with_progress_indicator(self):
        self.generate_snap_pair()
        base_delta = deltas.XDelta3Generator(
            source_path=self.source_file, target_path=self.target_file
        )

        message = "creating delta file from {!r}...".format(self.source_file)
        maxval = 10
        progress_indicator = ProgressBar(
            widgets=[message, AnimatedMarker()], maxval=maxval
        )
        progress_indicator.start()

        path = base_delta.make_delta(
            progress_indicator=progress_indicator, is_for_test=True
        )
        progress_indicator.finish()

        self.assertThat(path, m.FileExists())
        expect_path = "{}.{}".format(
            base_delta.target_path, base_delta.delta_file_extname
        )
        self.assertThat(path, m.Equals(expect_path))

    def test_xdelta3_with_custom_output_dir(self):
        self.generate_snap_pair()
        base_delta = deltas.XDelta3Generator(
            source_path=self.source_file, target_path=self.target_file
        )
        delta_filename = "{}.{}".format(
            os.path.split(base_delta.target_path)[1], base_delta.delta_file_extname
        )

        existed_output_dir = self.useFixture(fixtures.TempDir()).path
        path = base_delta.make_delta(existed_output_dir, is_for_test=True)

        expect_path = os.path.join(existed_output_dir, delta_filename)
        self.assertThat(path, m.FileExists())
        self.assertThat(path, m.Equals(expect_path))

        none_existed_output_dir = (
            self.useFixture(fixtures.TempDir()).path + "/whatever/"
        )
        path = base_delta.make_delta(none_existed_output_dir, is_for_test=True)

        expect_path = os.path.join(none_existed_output_dir, delta_filename)
        self.assertThat(path, m.FileExists())
        self.assertThat(path, m.Equals(expect_path))

    def test_xdelta3_logs(self):
        self.generate_snap_pair()
        base_delta = deltas.XDelta3Generator(
            source_path=self.source_file, target_path=self.target_file
        )
        base_delta.make_delta(is_for_test=True)

        self.assertThat(
            self.fake_logger.output,
            m.Contains(
                "Generating delta for {!r}".format(
                    os.path.basename(base_delta.target_path)
                )
            ),
        )
        self.assertThat(
            self.fake_logger.output, m.Contains("xdelta3 delta diff generation")
        )

    @mock.patch("subprocess.Popen")
    def test_xdelta3_return_invalid_code(self, mock_subproc_popen):
        # mock the subprocess.Popen with a unexpected returncode
        process_mock = mock.Mock()
        attrs = {"returncode": -1}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        self.generate_snap_pair()
        base_delta = deltas.XDelta3Generator(
            source_path=self.source_file, target_path=self.target_file
        )

        self.assertThat(
            lambda: base_delta.make_delta(is_for_test=True),
            m.raises(deltas.errors.DeltaGenerationError),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import json
import logging
import subprocess
from typing import Any, Dict, Iterator, Optional

from snapcraft.internal import repo

logger = logging.getLogger(__name__)


def _install_ua_tools() -> None:
    """Ensure UA tools are installed."""
    repo.Repo.install_build_packages(package_names=["ubuntu-advantage-tools"])


def _attach(ua_token: str) -> None:
    """Attach UA token (ua will update package lists)."""
    subprocess.check_call(["sudo", "ua", "attach", ua_token])


def _detach() -> None:
    """Detach UA token (ua will update package lists)."""
    subprocess.check_call(["sudo", "ua", "detach", "--assume-yes"])


def _is_attached() -> bool:
    return _status()["attached"]


def _status() -> Dict[str, Any]:
    stdout = subprocess.check_output(["sudo", "ua", "status", "--format", "json"])
    return json.loads(stdout)


@contextlib.contextmanager
def ua_manager(ua_token: Optional[str]) -> Iterator[None]:
    """Attach and detach UA token as required.

    Uses try/finally to ensure that token is detached on error.

    If ua_token is None, neither attach / detach will occur and the context
    manager is effectively a noop.

    If ua_token is specified and the system is already attached, log a warning
    that we are refusing to configure specified token, but otherwise resume
    normally.

    :param ua_token: Optional ua_token.
    """
    ua_needs_detach = False

    if ua_token is not None:
        # Make sure the tools are installed first.
        _install_ua_tools()

        if _is_attached():
            logger.warning(
                "Refusing to attach specified UA token as system is already attached. "
            )
        else:
            ua_needs_detach = True
            logger.info("Attaching specified UA token...")
            _attach(ua_token)

    try:
        yield
    finally:
        if ua_needs_detach:
            logger.info("Detaching specified UA token...")
            _detach()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import logging
import urllib3
from simplejson.scanner import JSONDecodeError

from snapcraft.internal.errors import SnapcraftError

logger = logging.getLogger(__name__)


_STORE_STATUS_URL = "https://status.snapcraft.io/"


# TODO: migrate to storeapi private exception to ready craft-store.
class HttpClientError(SnapcraftError):
    """Base class http client errors.

    :cvar fmt: A format string that daughter classes override
    """

    def __init__(self, **kwargs):
        with contextlib.suppress(KeyError, AttributeError):
            logger.debug("Store error response: {}".format(kwargs["response"].__dict__))
        super().__init__(**kwargs)


class StoreServerError(HttpClientError):

    fmt = "{what}: {error_text} (code {error_code}).\n{action}"

    def __init__(self, response):
        what = "The Snap Store encountered an error while processing your request"
        error_code = response.status_code
        error_text = response.reason
        action = "The operational status of the Snap Store can be checked at {}".format(
            _STORE_STATUS_URL
        )
        self.response = response

        super().__init__(
            response=response,
            what=what,
            error_text=error_text,
            error_code=error_code,
            action=action,
        )


class StoreNetworkError(HttpClientError):

    fmt = "There seems to be a network error: {message}"

    def __init__(self, exception):
        message = str(exception)
        with contextlib.suppress(IndexError):
            underlying_exception = exception.args[0]
            if isinstance(underlying_exception, urllib3.exceptions.MaxRetryError):
                message = (
                    "maximum retries exceeded trying to reach the store.\n"
                    "Check your network connection, and check the store "
                    "status at {}".format(_STORE_STATUS_URL)
                )
        super().__init__(message=message)


class InvalidCredentialsError(HttpClientError):

    fmt = 'Invalid credentials: {message}. Have you run "snapcraft login"?'

    def __init__(self, message):
        super().__init__(message=message)


class StoreAuthenticationError(HttpClientError):

    fmt = "Authentication error: {message}"

    def __init__(self, message, response=None):
        # Unfortunately the store doesn't give us a consistent error response,
        # so we'll check the ones of which we're aware.
        with contextlib.suppress(AttributeError, JSONDecodeError):
            response_json = response.json()
            extra_error_message = ""
            if "error_message" in response_json:
                extra_error_message = response_json["error_message"]
            elif "message" in response_json:
                extra_error_message = response_json["message"]

            if extra_error_message:
                message += ": {}".format(extra_error_message)

        super().__init__(response=response, message=message)


class StoreTwoFactorAuthenticationRequired(StoreAuthenticationError):
    def __init__(self):
        super().__init__("Two-factor authentication required.")


class InvalidLoginConfig(HttpClientError):

    fmt = "Invalid login config: {error}"

    def __init__(self, error):
        super().__init__(error=error)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import importlib.abc
import importlib.machinery
import importlib.util
import sys
import warnings

_VALID_V1_PLUGINS = [
    "ant",
    "catkin-tools",
    "conda",
    "dump",
    "gradle",
    "make",
    "nil",
    "python",
    "rust",
    "autotools",
    "cmake",
    "go",
    "kbuild",
    "maven",
    "nodejs",
    "qmake",
    "scons",
    "catkin",
    "colcon",
    "crystal",
    "dotnet",
    "godeps",
    "kernel",
    "meson",
    "plainbox-provider",
    "ruby",
    "waf",
]


class LegacyPluginLoader(importlib.abc.Loader):
    @classmethod
    def create_module(cls, spec):
        # Load the plugin from the new location.
        plugin_name = spec.name.split(".")[-1]
        return importlib.import_module(f"snapcraft.plugins.v1.{plugin_name}")

    @classmethod
    def exec_module(cls, module):
        # Rewrite the module __name__ to have that of the legacy import path.
        plugin_name = module.__name__.split(".")[-1]
        module.__name__ = f"snapcraft.plugins.{plugin_name}"
        return module


class LegacyPluginPathFinder(importlib.machinery.PathFinder):
    @classmethod
    def find_spec(cls, fullname, path=None, target=None):
        # Ensure plugins using their original import paths can be found and
        # warn about their new import path.
        if fullname in [f"snapcraft.plugins.{p}" for p in _VALID_V1_PLUGINS]:
            warnings.warn("Plugin import path has changed to 'snapcraft.plugins.v1'")
            return importlib.machinery.ModuleSpec(fullname, LegacyPluginLoader)
        else:
            return None


# https://docs.python.org/3/library/importlib.html#importlib.machinery.PathFinder
sys.meta_path.append(LegacyPluginPathFinder)  # type: ignore



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest.mock import patch

import fixtures
import progressbar
import requests

from snapcraft.internal import indicators
from tests import unit


class DumbTerminalTests(unit.TestCase):
    @patch("os.isatty")
    def setUp(self, mock_os_isatty):
        super().setUp()
        self.mock_os_isatty = mock_os_isatty
        self.mock_os_isatty.return_value = True

    def test_tty_terminal(self):
        self.assertTrue(indicators.is_dumb_terminal())

    def test_not_a_tty_terminal(self):
        self.mock_os_isatty.return_value = False
        self.assertFalse(indicators.is_dumb_terminal())

    def test_dumb_terminal_environment(self):
        self.useFixture(fixtures.EnvironmentVariable("TERM", "dumb"))
        self.assertTrue(indicators.is_dumb_terminal())

    def test_vt100_terminal_environmment(self):
        self.useFixture(fixtures.EnvironmentVariable("TERM", "vt100"))
        self.assertFalse(indicators.is_dumb_terminal())


class TestProgressBarInitialization:

    scenarios = [("Terminal", {"is_dumb": True}), ("Dumb Terminal", {"is_dumb": False})]

    def test_init_progress_bar_with_length(self, monkeypatch, is_dumb):
        monkeypatch.setattr(indicators, "is_dumb_terminal", lambda: is_dumb)

        pb = indicators._init_progress_bar(10, "destination", "message")

        assert pb.maxval == 10
        assert "message" in pb.widgets

        pb_widgets_types = [type(w) for w in pb.widgets]

        assert type(progressbar.Percentage()) in pb_widgets_types
        assert (type(progressbar.Bar()) in pb_widgets_types) is not is_dumb

    def test_init_progress_bar_with_unknown_length(self, monkeypatch, is_dumb):
        monkeypatch.setattr(indicators, "is_dumb_terminal", lambda: is_dumb)

        pb = indicators._init_progress_bar(0, "destination", "message")

        assert pb.maxval == progressbar.UnknownLength
        assert "message" in pb.widgets

        pb_widgets_types = [type(w) for w in pb.widgets]

        assert (type(progressbar.AnimatedMarker()) in pb_widgets_types) is not is_dumb


class IndicatorsDownloadTests(unit.FakeFileHTTPServerBasedTestCase):
    def setUp(self):
        super().setUp()

        dest_dir = "dst"
        os.makedirs(dest_dir)
        self.file_name = "snapcraft.yaml"
        self.dest_file = os.path.join(dest_dir, self.file_name)
        self.source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=self.file_name
        )

    def test_download_request_stream(self):
        request = requests.get(self.source, stream=True, allow_redirects=True)
        indicators.download_requests_stream(request, self.dest_file)

        self.assertTrue(os.path.exists(self.dest_file))

    def test_download_urllib_source(self):
        indicators.download_urllib_source(self.source, self.dest_file)

        self.assertTrue(os.path.exists(self.dest_file))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from snapcraft.internal.states import GlobalState

_scenarios = [
    (
        "build-snaps, build-packages and required-grade",
        dict(
            build_packages=["pkg1", "pkg2"],
            build_snaps=["snap1", "snap2"],
            required_grade="stable",
        ),
    ),
    (
        "build-snaps and no build-packages nor required-grade",
        dict(build_packages=[], build_snaps=["snap1", "snap2"], required_grade=None),
    ),
    (
        "no build-snaps, build-packages and no required-grade",
        dict(build_packages=["pkg1", "pkg2"], build_snaps=[], required_grade=None),
    ),
    (
        "no build-snaps nor build-packages and required-grade",
        dict(build_packages=[], build_snaps=[], required_grade="stable"),
    ),
]


class TestGlobalState:

    scenarios = _scenarios

    def test_save(self, tmp_work_path, build_packages, build_snaps, required_grade):
        global_state = GlobalState()
        global_state.append_build_packages(build_packages)
        global_state.append_build_snaps(build_snaps)
        global_state.set_required_grade(required_grade)

        global_state.save(filepath="state")

        prepend = "  - "

        if build_packages:
            build_packages = "\n" + "\n".join(
                ["{}{}".format(prepend, p) for p in build_packages]
            )
        else:
            build_packages = " []"

        if build_snaps:
            build_snaps = "\n" + "\n".join(
                ["{}{}".format(prepend, p) for p in build_snaps]
            )
        else:
            build_snaps = " []"

        if required_grade:
            required_grade = required_grade
        else:
            required_grade = "null"

        with open("state") as state_file:
            state_file_contents = state_file.read()
            assert (
                state_file_contents
                == dedent(
                    """\
                    !GlobalState
                    assets:
                      build-packages:{}
                      build-snaps:{}
                      required-grade: {}
                    """
                ).format(build_packages, build_snaps, required_grade)
            )

    def test_load(self, tmp_work_path, build_packages, build_snaps, required_grade):
        if required_grade:
            required_grade = required_grade
        else:
            required_grade = "null"

        with open("state", "w") as state_file:
            print(
                dedent(
                    """\
                !GlobalState
                assets:
                  build-packages: {}
                  build-snaps: {}
                  required-grade: {}
                """
                ).format(build_packages, build_snaps, required_grade),
                file=state_file,
            )

        global_state = GlobalState.load(filepath="state")

        assert global_state.get_build_packages() == build_packages
        assert global_state.get_build_snaps() == build_snaps

    def test_save_load_and_append(
        self, tmp_work_path, build_packages, build_snaps, required_grade
    ):
        global_state = GlobalState()
        global_state.append_build_packages(build_packages)
        global_state.append_build_snaps(build_snaps)
        global_state.save(filepath="state")

        assert global_state.get_build_packages() == build_packages
        assert global_state.get_build_snaps() == build_snaps

        new_packages = ["new-pkg1", "new-pkg2"]
        new_snaps = ["new-snap1", "new-snap2"]
        global_state = GlobalState.load(filepath="state")
        global_state.append_build_packages(new_packages)
        global_state.append_build_snaps(new_snaps)

        assert global_state.get_build_packages() == build_packages + new_packages

        assert global_state.get_build_snaps() == build_snaps + new_snaps

    def test_append_duplicate(
        self, tmp_work_path, build_packages, build_snaps, required_grade
    ):
        global_state = GlobalState()
        global_state.append_build_packages(build_packages)
        global_state.append_build_snaps(build_snaps)

        assert global_state.get_build_packages() == build_packages
        assert global_state.get_build_snaps() == build_snaps

        global_state.append_build_packages(build_packages)
        global_state.append_build_snaps(build_snaps)

        assert global_state.get_build_packages() == build_packages
        assert global_state.get_build_snaps() == build_snaps

    def test_load_with_missing(
        self, tmp_work_path, build_packages, build_snaps, required_grade
    ):
        with open("state", "w") as state_file:
            print("!GlobalState", file=state_file)
            print("assets: ", file=state_file)
            if build_packages:
                print("  build-packages: {}".format(build_packages), file=state_file)
            if build_snaps:
                print("  build-snaps: {}".format(build_snaps), file=state_file)
            if required_grade:
                print("  required-grade: {}".format(required_grade), file=state_file)

        global_state = GlobalState.load(filepath="state")

        assert global_state.get_build_packages() == build_packages
        assert global_state.get_build_snaps() == build_snaps
        assert global_state.get_required_grade() == required_grade



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from snapcraft.project import errors


class TestErrorFormatting:

    scenarios = [
        (
            "MissingSnapcraftYamlError",
            {
                "exception_class": errors.MissingSnapcraftYamlError,
                "kwargs": {"snapcraft_yaml_file_path": ".snapcraft.yaml"},
                "expected_message": (
                    "Could not find .snapcraft.yaml. Are you sure you are "
                    "in the right directory?\n"
                    "To start a new project, use `snapcraft init`"
                ),
            },
        ),
        (
            "DuplicateSnapcraftYamlError",
            {
                "exception_class": errors.DuplicateSnapcraftYamlError,
                "kwargs": {
                    "snapcraft_yaml_file_path": ".snapcraft.yaml",
                    "other_snapcraft_yaml_file_path": "snapcraft.yaml",
                },
                "expected_message": (
                    "Found a '.snapcraft.yaml' and a 'snapcraft.yaml'.\n"
                    "Please remove one and try again."
                ),
            },
        ),
    ]

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message


def test_SnapcraftExperimentalExtensionsRequiredError():
    error = errors.SnapcraftExperimentalExtensionsRequiredError(extension_name="foo")

    assert (
        error.get_brief()
        == "Experimental extension 'foo' is required, but not enabled."
    )
    assert error.get_details() is None
    assert (
        error.get_resolution()
        == "This extension may be enabled with the '--enable-experimental-extensions' parameter."
    )
    assert error.get_docs_url() is None
    assert error.get_exit_code() == 2



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
import re

import pytest

import snapcraft
import snapcraft.internal.project_loader.grammar._to as _to
from snapcraft.internal.project_loader import grammar


@pytest.mark.parametrize(
    "entry",
    [
        [{"on amd64,i386": ["foo"]}, {"on amd64,i386": ["bar"]}],
        [{"on amd64,i386": ["foo"]}, {"on i386,amd64": ["bar"]}],
    ],
)
def test_duplicates(entry):
    """Test that multiple identical selector sets is an error."""

    processor = grammar.GrammarProcessor(
        entry, snapcraft.ProjectOptions(), lambda x: True
    )
    with pytest.raises(grammar.errors.GrammarSyntaxError) as error:
        processor.process()

    expected = (
        "Invalid grammar syntax: found duplicate 'on amd64,i386' "
        "statements. These should be merged."
    )
    assert expected in str(error.value)


class TestBasicGrammar:

    scenarios = [
        (
            "unconditional",
            {
                "grammar_entry": ["foo", "bar"],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo", "bar"],
            },
        ),
        (
            "unconditional dict",
            {
                "grammar_entry": [{"foo": "bar"}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": [{"foo": "bar"}],
            },
        ),
        (
            "unconditional multi-dict",
            {
                "grammar_entry": [{"foo": "bar"}, {"foo2": "bar2"}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": [{"foo": "bar"}, {"foo2": "bar2"}],
            },
        ),
        (
            "mixed including",
            {
                "grammar_entry": ["foo", {"on i386": ["bar"]}],
                "host_arch": "i686",
                "target_arch": "i386",
                "expected_results": ["foo", "bar"],
            },
        ),
        (
            "mixed excluding",
            {
                "grammar_entry": ["foo", {"on i386": ["bar"]}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo"],
            },
        ),
        (
            "on amd64",
            {
                "grammar_entry": [{"on amd64": ["foo"]}, {"on i386": ["bar"]}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo"],
            },
        ),
        (
            "on i386",
            {
                "grammar_entry": [{"on amd64": ["foo"]}, {"on i386": ["bar"]}],
                "host_arch": "i686",
                "target_arch": "i386",
                "expected_results": ["bar"],
            },
        ),
        (
            "ignored else",
            {
                "grammar_entry": [{"on amd64": ["foo"]}, {"else": ["bar"]}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo"],
            },
        ),
        (
            "used else",
            {
                "grammar_entry": [{"on amd64": ["foo"]}, {"else": ["bar"]}],
                "host_arch": "i686",
                "target_arch": "i386",
                "expected_results": ["bar"],
            },
        ),
        (
            "nested amd64",
            {
                "grammar_entry": [
                    {"on amd64": [{"on amd64": ["foo"]}, {"on i386": ["bar"]}]}
                ],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo"],
            },
        ),
        (
            "nested amd64 dict",
            {
                "grammar_entry": [
                    {"on amd64": [{"on amd64": [{"foo": "bar"}]}, {"on i386": ["bar"]}]}
                ],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": [{"foo": "bar"}],
            },
        ),
        (
            "nested i386",
            {
                "grammar_entry": [
                    {"on i386": [{"on amd64": ["foo"]}, {"on i386": ["bar"]}]}
                ],
                "host_arch": "i686",
                "target_arch": "i386",
                "expected_results": ["bar"],
            },
        ),
        (
            "nested ignored else",
            {
                "grammar_entry": [
                    {"on amd64": [{"on amd64": ["foo"]}, {"else": ["bar"]}]}
                ],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo"],
            },
        ),
        (
            "nested used else",
            {
                "grammar_entry": [
                    {"on i386": [{"on amd64": ["foo"]}, {"else": ["bar"]}]}
                ],
                "host_arch": "i686",
                "target_arch": "amd64",
                "expected_results": ["bar"],
            },
        ),
        (
            "try",
            {
                "grammar_entry": [{"try": ["valid"]}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["valid"],
            },
        ),
        (
            "try else",
            {
                "grammar_entry": [{"try": ["invalid"]}, {"else": ["valid"]}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["valid"],
            },
        ),
        (
            "nested try",
            {
                "grammar_entry": [{"on amd64": [{"try": ["foo"]}, {"else": ["bar"]}]}],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": ["foo"],
            },
        ),
        (
            "nested try else",
            {
                "grammar_entry": [
                    {"on i386": [{"try": ["invalid"]}, {"else": ["bar"]}]}
                ],
                "host_arch": "i686",
                "target_arch": "i686",
                "expected_results": ["bar"],
            },
        ),
        (
            "optional",
            {
                "grammar_entry": ["foo", {"try": ["invalid"]}],
                "host_arch": "i686",
                "target_arch": "i386",
                "expected_results": ["foo"],
            },
        ),
        (
            "multi",
            {
                "grammar_entry": [
                    "foo",
                    {"on amd64": ["foo2"]},
                    {"on amd64 to arm64": ["foo3"]},
                ],
                "host_arch": "x86_64",
                "target_arch": "i386",
                "expected_results": ["foo", "foo2"],
            },
        ),
        (
            "multi-ordering",
            {
                "grammar_entry": [
                    "foo",
                    {"on amd64": ["on-foo"]},
                    "after-on",
                    {"on amd64 to i386": ["on-to-foo"]},
                    {"on amd64 to arm64": ["no-show"]},
                    "n-1",
                    "n",
                ],
                "host_arch": "x86_64",
                "target_arch": "i386",
                "expected_results": [
                    "foo",
                    "on-foo",
                    "after-on",
                    "on-to-foo",
                    "n-1",
                    "n",
                ],
            },
        ),
        (
            "complex nested dicts",
            {
                "grammar_entry": [
                    {"yes1": "yes1"},
                    {
                        "on amd64": [
                            {"yes2": "yes2"},
                            {"on amd64": [{"yes3": "yes3"}]},
                            {"yes4": "yes4"},
                            {"on i386": [{"no1": "no1"}]},
                            {"else": [{"yes5": "yes5"}]},
                            {"yes6": "yes6"},
                        ],
                    },
                    {"else": [{"no2": "no2"}]},
                    {"yes7": "yes7"},
                    {"on i386": [{"no3": "no3"}]},
                    {"else": [{"yes8": "yes8"}]},
                    {"yes9": "yes9"},
                ],
                "host_arch": "x86_64",
                "target_arch": "amd64",
                "expected_results": [
                    {"yes1": "yes1"},
                    {"yes2": "yes2"},
                    {"yes3": "yes3"},
                    {"yes4": "yes4"},
                    {"yes5": "yes5"},
                    {"yes6": "yes6"},
                    {"yes7": "yes7"},
                    {"yes8": "yes8"},
                    {"yes9": "yes9"},
                ],
            },
        ),
    ]

    def test_basic_grammar(
        self, monkeypatch, grammar_entry, host_arch, target_arch, expected_results
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        project = snapcraft.ProjectOptions(target_deb_arch=target_arch)

        processor = grammar.GrammarProcessor(
            grammar_entry, project, lambda x: "invalid" not in x
        )
        assert processor.process() == expected_results


class TestTransformerGrammar:

    scenarios = [
        (
            "unconditional",
            {
                "grammar_entry": ["foo", "bar"],
                "host_arch": "x86_64",
                "expected_results": ["foo", "bar"],
            },
        ),
        (
            "mixed including",
            {
                "grammar_entry": ["foo", {"on i386": ["bar"]}],
                "host_arch": "i686",
                "expected_results": ["foo", "bar"],
            },
        ),
        (
            "mixed excluding",
            {
                "grammar_entry": ["foo", {"on i386": ["bar"]}],
                "host_arch": "x86_64",
                "expected_results": ["foo"],
            },
        ),
        (
            "to",
            {
                "grammar_entry": [{"to i386": ["foo"]}],
                "host_arch": "x86_64",
                "expected_results": ["foo:i386"],
            },
        ),
        (
            "transform applies to nested",
            {
                "grammar_entry": [{"to i386": [{"on amd64": ["foo"]}]}],
                "host_arch": "x86_64",
                "expected_results": ["foo:i386"],
            },
        ),
        (
            "not to",
            {
                "grammar_entry": [{"to amd64": ["foo"]}, {"else": ["bar"]}],
                "host_arch": "x86_64",
                "expected_results": ["bar"],
            },
        ),
    ]

    def test_grammar_with_transformer(
        self, monkeypatch, grammar_entry, host_arch, expected_results
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        # Transform all 'to' statements to include arch
        def _transformer(call_stack, package_name, project_options):
            if any(isinstance(s, _to.ToStatement) for s in call_stack):
                if ":" not in package_name:
                    package_name += ":{}".format(project_options.deb_arch)

            return package_name

        processor = grammar.GrammarProcessor(
            grammar_entry,
            snapcraft.ProjectOptions(target_deb_arch="i386"),
            lambda x: True,
            transformer=_transformer,
        )

        assert processor.process() == expected_results


class TestInvalidGrammar:

    scenarios = [
        (
            "unmatched else",
            {
                "grammar_entry": [{"else": ["foo"]}],
                "expected_exception": ".*'else' doesn't seem to correspond.*",
            },
        ),
        (
            "unmatched else fail",
            {
                "grammar_entry": ["else fail"],
                "expected_exception": ".*'else' doesn't seem to correspond.*",
            },
        ),
        (
            "unexpected type",
            {
                "grammar_entry": [5],
                "expected_exception": ".*expected grammar section.*but got.*",
            },
        ),
    ]

    def test_invalid_grammar(self, grammar_entry, expected_exception):
        processor = grammar.GrammarProcessor(
            grammar_entry, snapcraft.ProjectOptions(), lambda x: True
        )

        with pytest.raises(grammar.errors.GrammarSyntaxError) as error:
            processor.process()

        assert re.match(expected_exception, str(error.value))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
from datetime import datetime
from textwrap import dedent

import pytest

from snapcraft.project import Project


def test_project_with_arguments():
    project = Project(target_deb_arch="armhf", debug=True)

    assert project.deb_arch == "armhf"
    assert project.debug is True

    # This is a backwards compatibility check
    assert project.info is None


def test_project_with_snapcraft_yaml_file_path_carries_info(tmp_work_path):
    snapcraft_yaml_path = pathlib.Path("snapcraft.yaml")
    with snapcraft_yaml_path.open("w") as snapcraft_yaml_file:
        print(
            dedent(
                """\
            name: foo
            version: "1"
            summary: bar
            description: baz
            confinement: strict

            parts:
              part1:
                plugin: go
            """
            ),
            file=snapcraft_yaml_file,
        )

    project = Project(snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix())

    # Only 1 value is enough
    assert project.info.name == "foo"


@pytest.mark.parametrize(
    "location", ["snap/snapcraft.yaml", "build-aux/snap/snapcraft.yaml"]
)
def test_project_local_plugin_location(tmp_work_path, location):
    snapcraft_yaml_path = tmp_work_path / pathlib.Path(location)
    snapcraft_yaml_path.parent.mkdir(parents=True)
    with snapcraft_yaml_path.open("w") as snapcraft_yaml_file:
        print(
            dedent(
                """\
            name: foo
            version: "1"
            summary: bar
            description: baz
            confinement: strict

            parts:
              part1:
                plugin: go
            """
            ),
            file=snapcraft_yaml_file,
        )

    project = Project(snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix())

    expected_plugins_dir = snapcraft_yaml_path.parent / "plugins"
    assert project.local_plugins_dir == expected_plugins_dir.as_posix()


def test_get_snapcraft_started():
    assert Project()._get_start_time() < datetime.utcnow()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import os
import pathlib
import tarfile
from textwrap import dedent
from unittest import mock
from xml.etree import ElementTree

import fixtures
import pytest
from testtools.matchers import Equals, FileExists, HasLength

from snapcraft.internal import errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.plugins.v1 import maven
from snapcraft.project import Project
from tests import unit

from . import PluginsV1BaseTestCase


class MavenPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = maven.MavenPlugin.schema()

        properties = schema["properties"]
        self.assertTrue(
            "maven-options" in properties,
            'Expected "maven-options" to be included in ' "properties",
        )

        maven_options = properties["maven-options"]

        self.assertTrue(
            "type" in maven_options, 'Expected "type" to be included in "maven-options"'
        )
        self.assertThat(
            maven_options["type"],
            Equals("array"),
            'Expected "maven-options" "type" to be "array", but '
            'it was "{}"'.format(maven_options["type"]),
        )

        self.assertTrue(
            "minitems" in maven_options,
            'Expected "minitems" to be included in "maven-options"',
        )
        self.assertThat(
            maven_options["minitems"],
            Equals(1),
            'Expected "maven-options" "minitems" to be 1, but '
            'it was "{}"'.format(maven_options["minitems"]),
        )

        self.assertTrue(
            "uniqueItems" in maven_options,
            'Expected "uniqueItems" to be included in "maven-options"',
        )
        self.assertTrue(
            maven_options["uniqueItems"],
            'Expected "maven-options" "uniqueItems" to be "True"',
        )

        maven_targets = properties["maven-targets"]

        self.assertTrue(
            "type" in maven_targets, 'Expected "type" to be included in "maven-targets"'
        )
        self.assertThat(
            maven_targets["type"],
            Equals("array"),
            'Expected "maven-targets" "type" to be "array", but '
            'it was "{}"'.format(maven_targets["type"]),
        )

        self.assertTrue(
            "minitems" in maven_targets,
            'Expected "minitems" to be included in "maven-targets"',
        )
        self.assertThat(
            maven_targets["minitems"],
            Equals(1),
            'Expected "maven-targets" "minitems" to be 1, but '
            'it was "{}"'.format(maven_targets["minitems"]),
        )

        self.assertTrue(
            "uniqueItems" in maven_targets,
            'Expected "uniqueItems" to be included in "maven-targets"',
        )
        self.assertTrue(
            maven_targets["uniqueItems"],
            'Expected "maven-targets" "uniqueItems" to be "True"',
        )

    def test_get_pull_properties(self):
        expected_pull_properties = [
            "maven-version",
            "maven-version-checksum",
            "maven-openjdk-version",
        ]
        resulting_pull_properties = maven.MavenPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["maven-options", "maven-targets"]
        resulting_build_properties = maven.MavenPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


def _get_expected_java_version(maven_plugin) -> str:
    base = maven_plugin.project._snap_meta.base
    maven_openjdk_version = maven_plugin.options.maven_openjdk_version

    if maven_openjdk_version:
        expected_java_version = maven_openjdk_version
    elif base in ("core", "core16"):
        expected_java_version = "9"
    else:
        expected_java_version = "11"

    return expected_java_version


_BASE_JAVA_COMBINATIONS = [
    ("", "core"),
    ("8", "core"),
    ("", "core16"),
    ("8", "core16"),
    ("", "core18"),
    ("11", "core18"),
]


@pytest.fixture(params=_BASE_JAVA_COMBINATIONS)
def maven_plugin(tmp_work_path, request):
    """Return an instance of MavenPlugin setup with different bases and java versions."""
    java_version, base = request.param

    class Options:
        maven_options = []
        maven_targets = [""]
        maven_version = maven._DEFAULT_MAVEN_VERSION
        maven_version_checksum = maven._DEFAULT_MAVEN_CHECKSUM
        maven_openjdk_version = java_version

    project = Project()
    project._snap_meta = Snap(name="test-snap", base=base, confinement="strict")

    return maven.MavenPlugin("test-part", Options(), project)


@pytest.fixture()
def maven_plugin_with_assets(maven_plugin):
    """Return an instance of MavenPlugin with artifacts setup using maven_plugin."""
    pathlib.Path(maven_plugin.sourcedir).mkdir(parents=True)

    expected_java_version = _get_expected_java_version(maven_plugin)

    fake_java_path = (
        pathlib.Path(maven_plugin.installdir)
        / f"usr/lib/jvm/java-{expected_java_version}-openjdk-amd64/bin/java"
    )
    fake_java_path.parent.mkdir(parents=True)
    fake_java_path.touch()

    maven_tar_path = (
        pathlib.Path(maven_plugin.partdir)
        / f"maven/apache-maven-{maven_plugin.options.maven_version}-bin.tar.gz"
    )
    maven_tar_path.parent.mkdir(parents=True)
    maven_tar_path.touch()

    return maven_plugin


def test_stage_and_build_packages(maven_plugin):
    expected_java_version = _get_expected_java_version(maven_plugin)

    assert maven_plugin.stage_packages == [
        f"openjdk-{expected_java_version}-jre-headless"
    ]
    assert maven_plugin.build_packages == [
        f"openjdk-{expected_java_version}-jdk-headless"
    ]


def test_build(mock_tar, mock_run, maven_plugin_with_assets):
    plugin = maven_plugin_with_assets

    def fake_run(cmd, *args, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "target"))
        open(os.path.join(plugin.builddir, "target", "jar.jar"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["mvn", "package"], cwd=plugin.builddir, env=mock.ANY
    )


def test_build_war(mock_tar, mock_run, maven_plugin_with_assets):
    plugin = maven_plugin_with_assets

    def fake_run(cmd, *args, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "target"))
        open(os.path.join(plugin.builddir, "target", "war.war"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["mvn", "package"], cwd=plugin.builddir, env=mock.ANY
    )


def test_build_with_targets(mock_tar, mock_run, maven_plugin_with_assets):
    maven_plugin_with_assets.options.maven_targets = ["child1", "child2"]
    plugin = maven_plugin_with_assets

    def fake_run(cmd, *args, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "child1", "target"))
        os.makedirs(os.path.join(plugin.builddir, "child2", "target"))
        open(
            os.path.join(plugin.builddir, "child1", "target", "child1.jar"), "w"
        ).close()
        open(
            os.path.join(plugin.builddir, "child2", "target", "child2.jar"), "w"
        ).close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["mvn", "package"], cwd=plugin.builddir, env=mock.ANY
    )


class TestBuildWithProxies(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "core18"

        class Options:
            maven_options = []
            maven_targets = [""]
            maven_version = maven._DEFAULT_MAVEN_VERSION
            maven_version_checksum = maven._DEFAULT_MAVEN_CHECKSUM
            maven_openjdk_version = "11"

        self.options = Options()

        patcher = mock.patch("snapcraft.internal.common.run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.sources.Tar")
        self.tar_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def create_assets(self, plugin):
        os.makedirs(plugin.sourcedir)

        fake_java_path = os.path.join(
            plugin.installdir,
            "usr",
            "lib",
            "jvm",
            "java-11-openjdk-amd64",
            "bin",
            "java",
        )
        os.makedirs(os.path.dirname(fake_java_path))
        open(fake_java_path, "w").close()

        maven_tar_path = os.path.join(
            plugin.partdir,
            "maven",
            "apache-maven-{}-bin.tar.gz".format(plugin.options.maven_version),
        )
        os.makedirs(os.path.dirname(maven_tar_path))
        tarfile.TarFile(maven_tar_path, "w").close()

    @staticmethod
    def _canonicalize_settings(settings):
        with io.StringIO(settings) as f:
            tree = ElementTree.parse(f)
        for element in tree.iter():
            if element.text is not None and element.text.isspace():
                element.text = None
            if element.tail is not None and element.tail.isspace():
                element.tail = None
        with io.StringIO() as f:
            tree.write(
                f,
                encoding="unicode",
                default_namespace="http://maven.apache.org/SETTINGS/1.0.0",
            )
            return f.getvalue() + "\n"

    def assertSettingsEqual(self, expected, actual_file):
        with open(actual_file) as fr:
            observed = fr.read()

        print(repr(self._canonicalize_settings(expected)))
        print(repr(self._canonicalize_settings(observed)))
        self.assertThat(
            self._canonicalize_settings(observed),
            Equals(self._canonicalize_settings(expected)),
        )

    def test_build_with_http_proxy(self):
        env_vars = (
            ("http_proxy", "http://localhost:3132"),
            ("https_proxy", None),
            ("no_proxy", None),
        )
        for v in env_vars:
            self.useFixture(fixtures.EnvironmentVariable(v[0], v[1]))

        plugin = maven.MavenPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        settings_path = os.path.join(plugin.partdir, "m2", "settings.xml")
        self.run_mock.assert_called_once_with(
            ["mvn", "package", "-s", settings_path], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertThat(settings_path, FileExists())
        expected_content = dedent(
            """\
            <settings xmlns="http://maven.apache.org/SETTINGS/1.0.0"
                      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                      xsi:schemaLocation="http://maven.apache.org/SETTINGS/1.0.0 http://maven.apache.org/xsd/settings-1.0.0.xsd">
              <interactiveMode>false</interactiveMode>
              <proxies>
                <proxy>
                  <id>http_proxy</id>
                  <active>true</active>
                  <protocol>http</protocol>
                  <host>localhost</host>
                  <port>3132</port>
                  <nonProxyHosts>localhost</nonProxyHosts>
                </proxy>
              </proxies>
            </settings>
        """
        )
        self.assertSettingsEqual(expected_content, settings_path)

    def test_build_with_http_proxy_and_no_proxy(self):
        env_vars = (
            ("http_proxy", "http://localhost:3132"),
            ("https_proxy", None),
            ("no_proxy", "internal"),
        )
        for v in env_vars:
            self.useFixture(fixtures.EnvironmentVariable(v[0], v[1]))

        plugin = maven.MavenPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        settings_path = os.path.join(plugin.partdir, "m2", "settings.xml")
        self.run_mock.assert_called_once_with(
            ["mvn", "package", "-s", settings_path], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertThat(settings_path, FileExists())
        expected_content = dedent(
            """\
            <settings xmlns="http://maven.apache.org/SETTINGS/1.0.0"
                      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                      xsi:schemaLocation="http://maven.apache.org/SETTINGS/1.0.0 http://maven.apache.org/xsd/settings-1.0.0.xsd">
              <interactiveMode>false</interactiveMode>
              <proxies>
                <proxy>
                  <id>http_proxy</id>
                  <active>true</active>
                  <protocol>http</protocol>
                  <host>localhost</host>
                  <port>3132</port>
                  <nonProxyHosts>internal</nonProxyHosts>
                </proxy>
              </proxies>
            </settings>
        """
        )
        self.assertSettingsEqual(expected_content, settings_path)

    def test_build_with_http_proxy_and_no_proxies(self):
        env_vars = (
            ("http_proxy", "http://localhost:3132"),
            ("https_proxy", None),
            ("no_proxy", "internal, pseudo-dmz"),
        )
        for v in env_vars:
            self.useFixture(fixtures.EnvironmentVariable(v[0], v[1]))

        plugin = maven.MavenPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        settings_path = os.path.join(plugin.partdir, "m2", "settings.xml")
        self.run_mock.assert_called_once_with(
            ["mvn", "package", "-s", settings_path], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertThat(settings_path, FileExists())
        expected_content = dedent(
            """\
            <settings xmlns="http://maven.apache.org/SETTINGS/1.0.0"
                      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                      xsi:schemaLocation="http://maven.apache.org/SETTINGS/1.0.0 http://maven.apache.org/xsd/settings-1.0.0.xsd">
              <interactiveMode>false</interactiveMode>
              <proxies>
                <proxy>
                  <id>http_proxy</id>
                  <active>true</active>
                  <protocol>http</protocol>
                  <host>localhost</host>
                  <port>3132</port>
                  <nonProxyHosts>internal|pseudo-dmz</nonProxyHosts>
                </proxy>
              </proxies>
            </settings>
            """
        )
        self.assertSettingsEqual(expected_content, settings_path)

    def test_build_with_http_and_https_proxy(self):
        env_vars = (
            ("http_proxy", "http://localhost:3132"),
            ("https_proxy", "http://localhost:3133"),
            ("no_proxy", None),
        )
        for v in env_vars:
            self.useFixture(fixtures.EnvironmentVariable(v[0], v[1]))

        plugin = maven.MavenPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        settings_path = os.path.join(plugin.partdir, "m2", "settings.xml")
        self.run_mock.assert_called_once_with(
            ["mvn", "package", "-s", settings_path], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertThat(settings_path, FileExists())
        expected_content = dedent(
            """\
            <settings xmlns="http://maven.apache.org/SETTINGS/1.0.0"
                      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                      xsi:schemaLocation="http://maven.apache.org/SETTINGS/1.0.0 http://maven.apache.org/xsd/settings-1.0.0.xsd">
              <interactiveMode>false</interactiveMode>
              <proxies>
                <proxy>
                  <id>http_proxy</id>
                  <active>true</active>
                  <protocol>http</protocol>
                  <host>localhost</host>
                  <port>3132</port>
                  <nonProxyHosts>localhost</nonProxyHosts>
                </proxy>
                <proxy>
                  <id>https_proxy</id>
                  <active>true</active>
                  <protocol>https</protocol>
                  <host>localhost</host>
                  <port>3133</port>
                  <nonProxyHosts>localhost</nonProxyHosts>
                </proxy>
              </proxies>
            </settings>
            """
        )
        self.assertSettingsEqual(expected_content, settings_path)

    def test_build_with_authenticated_proxies(self):
        env_vars = (
            ("http_proxy", "http://user1:pass1@localhost:3132"),
            ("https_proxy", "http://user2:pass2@localhost:3133"),
            ("no_proxy", None),
        )
        for v in env_vars:
            self.useFixture(fixtures.EnvironmentVariable(v[0], v[1]))

        plugin = maven.MavenPlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        def side(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "target"))
            open(os.path.join(plugin.builddir, "target", "dummy.jar"), "w").close()

        self.run_mock.side_effect = side

        plugin.build()

        settings_path = os.path.join(plugin.partdir, "m2", "settings.xml")
        self.run_mock.assert_called_once_with(
            ["mvn", "package", "-s", settings_path], cwd=plugin.builddir, env=mock.ANY
        )
        self.assertThat(settings_path, FileExists())
        expected_content = dedent(
            """\
            <settings xmlns="http://maven.apache.org/SETTINGS/1.0.0"
                      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                      xsi:schemaLocation="http://maven.apache.org/SETTINGS/1.0.0 http://maven.apache.org/xsd/settings-1.0.0.xsd">
              <interactiveMode>false</interactiveMode>
              <proxies>
                <proxy>
                  <id>http_proxy</id>
                  <active>true</active>
                  <protocol>http</protocol>
                  <host>localhost</host>
                  <port>3132</port>
                  <username>user1</username>
                  <password>pass1</password>
                  <nonProxyHosts>localhost</nonProxyHosts>
                </proxy>
                <proxy>
                  <id>https_proxy</id>
                  <active>true</active>
                  <protocol>https</protocol>
                  <host>localhost</host>
                  <port>3133</port>
                  <username>user2</username>
                  <password>pass2</password>
                  <nonProxyHosts>localhost</nonProxyHosts>
                </proxy>
              </proxies>
            </settings>
            """
        )
        self.assertSettingsEqual(expected_content, settings_path)


class MavenPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"
            maven_version = "3.3"
            maven_version_checksum = "sha1/1234567890"
            maven_openjdk_version = "10"

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            maven.MavenPlugin,
            "test-part",
            self.options,
            self.project,
        )


class TestUnsupportedJDKVersionError:

    scenarios = (
        (
            "core",
            dict(
                base="core",
                version="11",
                expected_message=(
                    "The maven-openjdk-version plugin property was set to '11'.\n"
                    "Valid values for the 'core' base are: '8' or '9'."
                ),
            ),
        ),
        (
            "core16",
            dict(
                base="core16",
                version="11",
                expected_message=(
                    "The maven-openjdk-version plugin property was set to '11'.\n"
                    "Valid values for the 'core16' base are: '8' or '9'."
                ),
            ),
        ),
        (
            "core18",
            dict(
                base="core18",
                version="9",
                expected_message=(
                    "The maven-openjdk-version plugin property was set to '9'.\n"
                    "Valid values for the 'core18' base are: '11' or '8'."
                ),
            ),
        ),
    )

    def test_use_invalid_openjdk_version_fails(self, base, version, expected_message):
        class Options:
            maven_options = []
            maven_targets = [""]
            maven_version = maven._DEFAULT_MAVEN_VERSION
            maven_version_checksum = maven._DEFAULT_MAVEN_CHECKSUM
            maven_openjdk_version = version

        project = Project()
        project._snap_meta = Snap(name="test-snap", base=base, confinement="strict")

        with pytest.raises(maven.UnsupportedJDKVersionError) as error:
            maven.MavenPlugin("test-part", Options(), project)
            assert str(error) == expected_message



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import distutils.util
import glob
import os
import shutil
import subprocess
import sys
from distutils import dir_util
from textwrap import dedent
from typing import Callable, List, Union

import fixtures
import requests
import testtools
import yaml as yaml_utils
from pexpect import popen_spawn
from testtools import content

from tests import fixture_setup, os_release, subprocess_utils
from tests.integration import platform


class RegisterError(Exception):
    pass


class TestCase(testtools.TestCase):
    def setUp(self):
        super().setUp()

        self.patchelf_command = "patchelf"
        self.execstack_command = "execstack"

        package_type = os.getenv("SNAPCRAFT_PACKAGE_TYPE")
        if package_type == "snap":
            self.snapcraft_command = "/snap/bin/snapcraft"
            self.patchelf_command = "/snap/snapcraft/current/bin/patchelf"
            self.execstack_command = "/snap/snapcraft/current/usr/bin/execstack"
        elif package_type == "deb":
            self.snapcraft_command = "/usr/bin/snapcraft"
            self.snapcraft_parser_command = "/usr/bin/snapcraft-parser"
        elif os.getenv("VIRTUAL_ENV") and sys.platform == "win32":
            self.snapcraft_command = os.path.join(
                os.getenv("VIRTUAL_ENV"), "Scripts", "snapcraft.exe"
            )
            self.snapcraft_parser_command = os.path.join(
                os.getenv("VIRTUAL_ENV"), "Scripts", "snapcraft-parser.exe"
            )
        elif os.getenv("VIRTUAL_ENV"):
            self.snapcraft_command = os.path.join(
                os.getenv("VIRTUAL_ENV"), "bin", "snapcraft"
            )
            self.snapcraft_parser_command = os.path.join(
                os.getenv("VIRTUAL_ENV"), "bin", "snapcraft-parser"
            )
        elif package_type == "brew":
            self.snapcraft_command = "/usr/local/bin/snapcraft"
        else:
            raise EnvironmentError(
                "snapcraft is not setup correctly for testing. Either set "
                "SNAPCRAFT_PACKAGE_TYPE to 'snap', 'deb' or 'brew', to run from "
                "either the snap, deb or homebrew or make sure your venv is properly "
                "setup as described in HACKING.md."
            )

        self.snaps_dir = os.path.join(os.path.dirname(__file__), "snaps")
        temp_cwd_fixture = fixture_setup.TempCWD()
        self.useFixture(temp_cwd_fixture)
        self.path = temp_cwd_fixture.path

        # Use a separate path for XDG dirs, or changes there may be detected as
        # source changes.
        self.xdg_path = self.useFixture(fixtures.TempDir()).path
        self.useFixture(fixture_setup.TempXDG(self.xdg_path))

        # Use this host to run through the lifecycle tests
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )

        # Use a dumb terminal for tests
        self.useFixture(fixtures.EnvironmentVariable("TERM", "dumb"))

        # Force appearance of TTY for pexpect to work with echo.is_connected_tty()
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_HAS_TTY", "y"))

        # Disable Sentry reporting for tests, otherwise they'll hang waiting
        # for input
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_ERROR_REPORTING", "false")
        )

        # Don't let the managed host variable leak into tests
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_MANAGED_HOST"))

        # Note that these directories won't exist when the test starts,
        # they might be created after calling the snapcraft command on the
        # project dir.
        self.parts_dir = "parts"
        self.stage_dir = "stage"
        self.prime_dir = "prime"

        self.deb_arch = platform.get_deb_arch()
        self.arch_triplet = platform.get_arch_triplet()

        self.distro_series = os_release.get_version_codename()

    def run_snapcraft(
        self,
        command: Union[str, List[str]] = None,
        project_dir: str = None,
        debug: bool = True,
        pre_func: Callable[[], None] = lambda: None,
        env=None,
    ) -> str:
        if project_dir:
            self.copy_project_to_cwd(project_dir)

        if command is None:
            command = []
        if isinstance(command, str):
            command = [command]
        snapcraft_command = self.snapcraft_command
        if isinstance(snapcraft_command, str):
            snapcraft_command = [snapcraft_command]
        if debug:
            snapcraft_command.append("-d")
        try:
            pre_func()
            snapcraft_output = subprocess.check_output(
                snapcraft_command + command,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                env=env,
            )
        except subprocess.CalledProcessError as e:
            self.addDetail("command", content.text_content(str(self.snapcraft_command)))
            self.addDetail("output", content.text_content(e.output))
            raise
        except FileNotFoundError:
            self.addDetail("command", content.text_content(str(self.snapcraft_command)))
            raise

        if not os.getenv("SNAPCRAFT_IGNORE_APT_AUTOREMOVE", False):
            self.addCleanup(self.run_apt_autoremove)

        return snapcraft_output

    def spawn_snapcraft(self, command: Union[str, List[str]]):
        snapcraft_command = self.snapcraft_command
        if isinstance(snapcraft_command, str):
            snapcraft_command = [snapcraft_command]
        try:
            return popen_spawn.PopenSpawn(" ".join(snapcraft_command + command))
        except FileNotFoundError:
            self.addDetail("command", content.text_content(str(snapcraft_command)))

    def run_snapcraft_parser(self, arguments):
        try:
            snapcraft_output = subprocess.check_output(
                [self.snapcraft_parser_command, "-d"] + arguments,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
        except subprocess.CalledProcessError as e:
            self.addDetail("output", content.text_content(e.output))
            raise
        return snapcraft_output

    def run_apt_autoremove(self):
        if sys.platform == "win32":
            return

        deb_env = os.environ.copy()
        deb_env.update(
            {"DEBIAN_FRONTEND": "noninteractive", "DEBCONF_NONINTERACTIVE_SEEN": "true"}
        )

        try:
            autoremove_output = subprocess.check_output(
                "sudo apt-get autoremove -y".split(),
                stderr=subprocess.STDOUT,
                env=deb_env,
            )
            self.addDetail(
                "apt-get autoremove output",
                content.text_content(autoremove_output.decode("utf-8")),
            )
        except FileNotFoundError as e:
            self.addDetail("apt-get autoremove error", content.text_content(str(e)))
        except subprocess.CalledProcessError as e:
            self.addDetail("apt-get autoremove error", content.text_content(str(e)))
            self.addDetail(
                "apt-get autoremove output",
                content.text_content(e.output.decode("utf-8")),
            )

            if os.getenv("SNAPCRAFT_APT_AUTOREMOVE_CHECK_FAIL", False):
                raise

    def copy_project_to_cwd(self, project_dir: str) -> None:
        # Because cwd already exists, shutil.copytree would raise
        # FileExistsError. Use the lesser known distutils.dir_util.copy_tree
        dir_util.copy_tree(
            os.path.join(self.snaps_dir, project_dir), self.path, preserve_symlinks=True
        )

    def construct_yaml(
        self,
        name="test",
        version="0.1",
        summary="Simple test snap",
        description="Something something",
        grade=None,
        architectures=None,
        parts=dedent(
            """\
                           my-part:
                             plugin: nil
                           """
        ),
        build_packages="[]",
        adopt_info=None,
    ):
        snapcraft_yaml = {
            "name": name,
            "summary": summary,
            "description": description,
            "parts": yaml_utils.load(parts),
            "build-packages": yaml_utils.load(build_packages),
        }

        if version:
            snapcraft_yaml["version"] = version
        if adopt_info:
            snapcraft_yaml["adopt-info"] = adopt_info
        if grade:
            snapcraft_yaml["grade"] = grade
        if architectures:
            snapcraft_yaml["architectures"] = architectures

        with open("snapcraft.yaml", "w") as f:
            yaml_utils.dump(snapcraft_yaml, stream=f)

    def get_output_ignoring_non_zero_exit(self, binary, cwd=None):
        # Executing the binaries exists > 0 on trusty.
        # TODO investigate more to understand the cause.
        try:
            output = subprocess.check_output(binary, universal_newlines=True, cwd=cwd)
        except subprocess.CalledProcessError as exception:
            output = exception.output
        return output

    def set_stage_package_version(
        self, snapcraft_yaml_path, part, package, version=None
    ):
        return self.set_package_version(
            "stage-packages", snapcraft_yaml_path, part, package, version
        )

    def set_build_package_version(
        self, snapcraft_yaml_path, part, package, version=None
    ):
        return self.set_package_version(
            "build-packages", snapcraft_yaml_path, part, package, version
        )

    def set_package_version(
        self, type_, snapcraft_yaml_path, part, package, version=None
    ):
        # This doesn't handle complex package syntax.
        with open(snapcraft_yaml_path) as snapcraft_yaml_file:
            snapcraft_yaml = yaml_utils.load(snapcraft_yaml_file)
        if part:
            packages = snapcraft_yaml["parts"][part].get(type_, [])
        else:
            packages = snapcraft_yaml.get(type_, [])
        for index, package_in_yaml in enumerate(packages):
            if package_in_yaml.split("=")[0] == package:
                if version is None:
                    version = get_package_version(
                        package, self.distro_series, self.deb_arch
                    )

                packages[index] = "{}={}".format(package, version)
                break
        else:
            self.fail("The part {} doesn't have a package {}".format(part, package))

        with open(snapcraft_yaml_path, "w") as snapcraft_yaml_file:
            yaml_utils.dump(snapcraft_yaml, stream=snapcraft_yaml_file)
        return version

    def set_build_package_architecture(
        self, snapcraft_yaml_path, part, package, architecture
    ):
        # This doesn't handle complex package syntax.
        with open(snapcraft_yaml_path) as snapcraft_yaml_file:
            snapcraft_yaml = yaml_utils.load(snapcraft_yaml_file)
        packages = snapcraft_yaml["parts"][part]["build-packages"]
        for index, package_in_yaml in enumerate(packages):
            if package_in_yaml == package:
                packages[index] = "{}:{}".format(package, architecture)
                break
        else:
            self.fail("The part {} doesn't have a package {}".format(part, package))

        with open(snapcraft_yaml_path, "w") as snapcraft_yaml_file:
            yaml_utils.dump(snapcraft_yaml, stream=snapcraft_yaml_file)


class BzrSourceBaseTestCase(TestCase):
    def setUp(self):
        super().setUp()
        if shutil.which("bzr") is None:
            self.skipTest("bzr is not installed")

    def init_source_control(self):
        subprocess.check_call(
            ["bzr", "init", "."], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        subprocess.check_call(
            ["bzr", "whoami", "--branch", '"Example Dev <dev@example.com>"']
        )

    def commit(self, message, unchanged=False):
        command = ["bzr", "commit", "-m", message]
        if unchanged:
            command.append("--unchanged")
        subprocess.check_call(
            command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )

    def get_revno(self, path=None):
        command = ["bzr", "revno", "-r", "-1"]
        if path:
            command.append(path)
        return subprocess.check_output(command, universal_newlines=True).strip()


class GitSourceBaseTestCase(TestCase):
    def setUp(self):
        super().setUp()
        if shutil.which("git") is None:
            self.skipTest("git is not installed")

    def init_source_control(self):
        subprocess.check_call(["git", "init", "."], stdout=subprocess.DEVNULL)
        subprocess.check_call(
            ["git", "config", "--local", "user.name", '"Example Dev"']
        )
        subprocess.check_call(
            ["git", "config", "--local", "user.email", "dev@example.com"]
        )

    def add_file(self, file_path):
        subprocess.check_call(["git", "add", file_path], stdout=subprocess.DEVNULL)

    def commit(self, message, allow_empty=False):
        command = ["git", "commit", "-m", message]
        if allow_empty:
            command.append("--allow-empty")
        subprocess.check_call(command, stdout=subprocess.DEVNULL)

    def tag(self, tag_name):
        subprocess.check_call(
            ["git", "tag", "-a", "-m", tag_name, tag_name], stdout=subprocess.DEVNULL
        )

    def get_revno(self):
        return subprocess_utils.call_with_output(
            ["git", "rev-list", "HEAD", "--max-count=1"]
        )


class HgSourceBaseTestCase(TestCase):
    def setUp(self):
        super().setUp()
        if shutil.which("hg") is None:
            self.skipTest("mercurial is not installed")

    def init_source_control(self):
        subprocess.check_call(["hg", "init", "."])

    def commit(self, message, file_):
        subprocess.check_call(
            ["hg", "commit", "-m", message, "--user", '"Example Dev"', "-A", file_]
        )

    def get_revno(self, path=None):
        command = ["hg", "log", "--template", '"{desc}"', "-r", "-1"]
        if path:
            command.extend(["--cwd", path])
        return subprocess.check_output(command, universal_newlines=True).strip()

    def get_id(self):
        return subprocess_utils.call_with_output(["hg", "id"]).split()[0]


class SubversionSourceBaseTestCase(TestCase):
    def setUp(self):
        super().setUp()
        if shutil.which("svn") is None:
            self.skipTest("svn is not installed")

    def init_source_control(self):
        subprocess.check_call(["svnadmin", "create", "repo"], stdout=subprocess.DEVNULL)

    def checkout(self, source, destination):
        subprocess.check_call(
            ["svn", "checkout", source, destination], stdout=subprocess.DEVNULL
        )

    def add(self, file_path, cwd=None):
        subprocess.check_call(
            ["svn", "add", file_path], stdout=subprocess.DEVNULL, cwd=cwd
        )

    def commit(self, message, cwd=None):
        subprocess.check_call(
            ["svn", "commit", "-m", message], stdout=subprocess.DEVNULL, cwd=cwd
        )

    def update(self, cwd=None):
        subprocess.check_call(["svn", "update"], stdout=subprocess.DEVNULL, cwd=cwd)


class SnapdIntegrationTestCase(TestCase):

    slow_test = False

    def setUp(self) -> None:
        super().setUp()
        run_slow_tests = os.environ.get("SNAPCRAFT_SLOW_TESTS", False)
        if run_slow_tests:
            run_slow_tests = distutils.util.strtobool(str(run_slow_tests))
        if self.slow_test and not run_slow_tests:
            self.skipTest("Not running slow tests")
        if os.environ.get("ADT_TEST") and self.deb_arch == "armhf":
            self.skipTest("The autopkgtest armhf runners can't install snaps")

    def install_snap(self) -> None:
        try:
            subprocess.check_output(
                ["sudo", "snap", "install", glob.glob("*.snap")[0], "--dangerous"],
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
        except subprocess.CalledProcessError as e:
            self.addDetail("output", content.text_content(e.output))
            raise


def get_package_version(package_name, series, deb_arch):
    # http://people.canonical.com/~ubuntu-archive/madison.cgi?package=hello&a=amd64&c=&s=zesty&text=on
    params = {
        "package": package_name,
        "s": "{0},{0}-updates,{0}-security".format(series),
        "a": deb_arch,
        "text": "on",
    }
    query = requests.get(
        "https://people.canonical.com/~ubuntu-archive/madison.cgi", params
    )
    query.raise_for_status()
    package = query.text.strip().split("\n")[-1]
    package_status = [i.strip() for i in package.strip().split("|")]
    return package_status[1]


def add_stage_packages(
    *, part_name: str, stage_packages: List[str], snapcraft_yaml_file=None
):
    if snapcraft_yaml_file is None:
        snapcraft_yaml_file = os.path.join("snap", "snapcraft.yaml")

    with open(snapcraft_yaml_file) as file_read:
        y = yaml_utils.load(file_read)
        if "stage-packages" in y["parts"][part_name]:
            y["parts"][part_name]["stage-packages"].extend(stage_packages)
        else:
            y["parts"][part_name]["stage-packages"] = stage_packages
    with open(snapcraft_yaml_file, "w") as file_write:
        yaml_utils.dump(y, stream=file_write)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import os
import pathlib
from typing import Optional, TextIO
from urllib.parse import urljoin, urlparse

import pymacaroons
import requests
from simplejson.scanner import JSONDecodeError
from xdg import BaseDirectory

from . import agent, _config, errors, _http_client


UBUNTU_ONE_SSO_URL = "https://login.ubuntu.com/"


logger = logging.getLogger(__name__)


def _deserialize_macaroon(value):
    try:
        return pymacaroons.Macaroon.deserialize(value)
    except:  # noqa LP: #1733004
        raise errors.InvalidCredentialsError("Failed to deserialize macaroon")


def _macaroon_auth(conf):
    """Format a macaroon and its associated discharge.

    :return: A string suitable to use in an Authorization header.

    """
    root_macaroon_raw = conf.get("macaroon")
    if root_macaroon_raw is None:
        raise errors.InvalidCredentialsError("Root macaroon not in the config file")
    unbound_raw = conf.get("unbound_discharge")
    if unbound_raw is None:
        raise errors.InvalidCredentialsError("Unbound discharge not in the config file")

    root_macaroon = _deserialize_macaroon(root_macaroon_raw)
    unbound = _deserialize_macaroon(unbound_raw)
    bound = root_macaroon.prepare_for_request(unbound)
    discharge_macaroon_raw = bound.serialize()
    auth = "Macaroon root={}, discharge={}".format(
        root_macaroon_raw, discharge_macaroon_raw
    )

    return auth


class UbuntuOneSSOConfig(_config.Config):
    """Hold configuration options in sections.

    There can be two sections for the sso related credentials: production and
    staging. This is governed by the UBUNTU_ONE_SSO_URL environment
    variable. Other sections are ignored but preserved.

    """

    def _get_section_name(self) -> str:
        url = os.getenv("UBUNTU_ONE_SSO_URL", UBUNTU_ONE_SSO_URL)
        return urlparse(url).netloc

    def _get_config_path(self) -> pathlib.Path:
        return (
            pathlib.Path(BaseDirectory.save_config_path("snapcraft")) / "snapcraft.cfg"
        )


class UbuntuOneAuthClient(_http_client.Client):
    """Store Client using Ubuntu One SSO provided macaroons."""

    @staticmethod
    def _is_needs_refresh_response(response):
        return (
            response.status_code == requests.codes.unauthorized
            and response.headers.get("WWW-Authenticate") == "Macaroon needs_refresh=1"
        )

    def __init__(self, *, user_agent: str = agent.get_user_agent()) -> None:
        super().__init__(user_agent=user_agent)

        self._conf = UbuntuOneSSOConfig()
        self.auth_url = os.environ.get("UBUNTU_ONE_SSO_URL", UBUNTU_ONE_SSO_URL)

        try:
            self.auth: Optional[str] = _macaroon_auth(self._conf)
        except errors.InvalidCredentialsError:
            self.auth = None

    def _extract_caveat_id(self, root_macaroon):
        macaroon = pymacaroons.Macaroon.deserialize(root_macaroon)
        # macaroons are all bytes, never strings
        sso_host = urlparse(self.auth_url).netloc
        for caveat in macaroon.caveats:
            if caveat.location == sso_host:
                return caveat.caveat_id
        else:
            raise errors.InvalidCredentialsError("Invalid root macaroon")

    def login(
        self,
        *,
        email: Optional[str] = None,
        password: Optional[str] = None,
        macaroon: Optional[str] = None,
        otp: Optional[str] = None,
        config_fd: TextIO = None,
        save: bool = True,
    ) -> None:
        if config_fd is not None:
            self._conf.load(config_fd=config_fd)
        # Verbose to keep static checks happy.
        elif email is not None and password is not None and macaroon is not None:
            # Ask the store for the needed capabilities to be associated with
            # the macaroon.
            caveat_id = self._extract_caveat_id(macaroon)
            unbound_discharge = self._discharge_token(email, password, otp, caveat_id)
            # Clear any old data before setting.
            self._conf.clear()
            # The macaroon has been discharged, save it in the config
            self._conf.set("macaroon", macaroon)
            self._conf.set("unbound_discharge", unbound_discharge)
            self._conf.set("email", email)
        else:
            raise RuntimeError("Logic Error")

        # Set auth and headers.
        self.auth = _macaroon_auth(self._conf)

        if save:
            self._conf.save()

    def export_login(self, *, config_fd: TextIO, encode: bool = False) -> None:
        self._conf.save(config_fd=config_fd, encode=encode)

    def logout(self) -> None:
        self._conf.clear()
        self._conf.save()

    def _discharge_token(
        self, email: str, password: str, otp: Optional[str], caveat_id
    ) -> str:
        data = dict(email=email, password=password, caveat_id=caveat_id)
        if otp:
            data["otp"] = otp

        url = urljoin(self.auth_url, "/api/v2/tokens/discharge")

        response = self.request(
            "POST",
            url,
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if response.ok:
            return response.json()["discharge_macaroon"]

        try:
            response_json = response.json()
        except JSONDecodeError:
            response_json = dict()

        if response.status_code == requests.codes.unauthorized and any(
            error.get("code") == "twofactor-required"
            for error in response_json.get("error_list", [])
        ):
            raise errors.StoreTwoFactorAuthenticationRequired()
        else:
            raise errors.StoreAuthenticationError(
                "Failed to get unbound discharge", response
            )

    def _refresh_token(self, unbound_discharge):
        data = {"discharge_macaroon": unbound_discharge}
        url = urljoin(self.auth_url, "/api/v2/tokens/refresh")
        response = self.request(
            "POST",
            url,
            json=data,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if response.ok:
            return response.json()["discharge_macaroon"]
        else:
            raise errors.StoreAuthenticationError(
                "Failed to refresh unbound discharge", response
            )

    def request(
        self, method, url, params=None, headers=None, auth_header=True, **kwargs
    ) -> requests.Response:
        if headers and auth_header:
            headers["Authorization"] = self.auth
        elif auth_header:
            headers = {"Authorization": self.auth}

        response = super().request(
            method, url, params=params, headers=headers, **kwargs
        )

        if self._is_needs_refresh_response(response):
            unbound_discharge = self._refresh_token(self._conf.get("unbound_discharge"))
            self._conf.set("unbound_discharge", unbound_discharge)
            self._conf.save()
            self.auth = _macaroon_auth(self._conf)
            headers["Authorization"] = self.auth

            response = super().request(
                method, url, params=params, headers=headers, **kwargs
            )

        return response



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft import formatting_utils
from snapcraft.internal import steps


class OutdatedReport:
    """The OutdatedReport class explains why a given step is outdated.

    An outdated step is defined to be a step that has run, but since doing so
    one of the following things have happened:

    - A step earlier in the lifecycle has run again.

    - The source on disk has been updated.
    """

    def __init__(
        self, *, previous_step_modified: steps.Step = None, source_updated: bool = False
    ) -> None:
        """Create a new OutdatedReport.

        :param steps.step previous_step_modified: Step earlier in the lifecycle
                                                  that has changed.
        :param bool source_updated: Whether or not the source changed on disk.
        """
        self.previous_step_modified = previous_step_modified
        self.source_updated = source_updated

    def get_report(self) -> str:
        """Get verbose report.

        :return: Report why the step is outdated.
        :rtype: str
        """
        messages = []

        if self.previous_step_modified:
            messages.append(
                "The {!r} step has run more recently.\n".format(
                    self.previous_step_modified.name
                )
            )

        if self.source_updated:
            messages.append("The source has changed on disk.\n")

        return "".join(messages)

    def get_summary(self) -> str:
        """Get summarized report.

        :return: Short summary of why the step is dirty.
        :rtype: str
        """
        reasons = []

        if self.previous_step_modified:
            reasons.append("{!r} step".format(self.previous_step_modified.name))

        if self.source_updated:
            reasons.append("source")

        return "{} changed".format(formatting_utils.humanize_list(reasons, "and", "{}"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import xdg
from testtools.matchers import Equals

from snapcraft import config
from snapcraft.internal.errors import SnapcraftInvalidCLIConfigError
from tests import unit


class TestCLIConfig(unit.TestCase):
    def test_non_existing_file_succeeds(self):
        conf = config.CLIConfig()
        self.assertThat(conf.parser.sections(), Equals([]))

    def test_set_and_get_sentry_send_always_with_contextmanager(self):
        with config.CLIConfig() as cli_config:
            cli_config.set_sentry_send_always(True)

        with config.CLIConfig() as cli_config:
            self.assertThat(cli_config.get_sentry_send_always(), Equals(True))

    def test_set_and_get_sentry_send_always(self):
        cli_config = config.CLIConfig()
        cli_config.set_sentry_send_always(True)
        cli_config.save()

        new_cli_config = config.CLIConfig()
        new_cli_config.load()
        self.assertThat(new_cli_config.get_sentry_send_always(), Equals(True))

    def test_set_and_get_outdated_step_action_with_contextmanager(self):
        with config.CLIConfig() as cli_config:
            cli_config.set_outdated_step_action(config.OutdatedStepAction.CLEAN)

        with config.CLIConfig() as cli_config:
            self.assertThat(
                cli_config.get_outdated_step_action(),
                Equals(config.OutdatedStepAction.CLEAN),
            )

    def test_set_and_get_outdated_step_action(self):
        cli_config = config.CLIConfig()
        cli_config.set_outdated_step_action(config.OutdatedStepAction.CLEAN)
        cli_config.save()

        new_cli_config = config.CLIConfig()
        new_cli_config.load()
        self.assertThat(
            new_cli_config.get_outdated_step_action(),
            Equals(config.OutdatedStepAction.CLEAN),
        )

    def test_set_when_read_only(self):
        cli_config = config.CLIConfig(read_only=True)

        self.assertRaises(RuntimeError, cli_config.set_sentry_send_always, True)

    def test_save_when_read_only(self):
        cli_config = config.CLIConfig(read_only=True)

        self.assertRaises(RuntimeError, cli_config.save)

    def test_contextmanager_with_read_only(self):
        with config.CLIConfig(read_only=True) as cli_config:
            # This should be False
            self.assertThat(cli_config.get_sentry_send_always(), Equals(False))

    def test_load_invalid_config(self):
        config_path = os.path.join(
            xdg.BaseDirectory.save_config_path("snapcraft"), "cli.cfg"
        )
        with open(config_path, "w") as f:
            f.write("invalid config")

        cli_config = config.CLIConfig()

        self.assertRaises(SnapcraftInvalidCLIConfigError, cli_config.load)
